from typing import Sequence, Union

import numpy as np
from arbol import aprint, asection

from dexp.datasets import BaseDataset
from dexp.processing.remove_beads import BeadsRemover
from dexp.utils.backends import BestBackend
from dexp.utils.slicing import slice_from_shape


def dataset_extract_psf(
    dataset: BaseDataset,
    dest_path: str,
    channels: Sequence[str],
    slicing: Union[Sequence[slice], slice],
    peak_threshold: int = 500,
    similarity_threshold: float = 0.5,
    psf_size: int = 35,
    device: int = 0,
    stop_at_exception: bool = True,
    verbose: bool = True,
) -> None:
    """
    Computes PSF from beads.
    Additional information at dexp.processing.remove_beads.beadremover documentation.
    """
    dest_path = dest_path.split(".")[0]

    remove_beads = BeadsRemover(
        peak_threshold=peak_threshold, similarity_threshold=similarity_threshold, psf_size=psf_size, verbose=verbose
    )

    for channel in dataset._selected_channels(channels):
        array = dataset.get_array(channel)
        _, volume_slicing, time_points = slice_from_shape(array.shape, slicing)

        psfs = []
        for i in range(len(time_points)):
            tp = time_points[i]
            try:
                with asection(f"Removing beads of channel: {channel}"):
                    with asection(f"Loading time point {i}/{len(time_points)}"):
                        tp_array = np.asarray(array[tp][volume_slicing])

                    with asection("Processing"):
                        with BestBackend(exclusive=True, enable_unified_memory=True, device_id=device) as backend:
                            tp_array = backend.to_backend(tp_array)
                            estimated_psf = remove_beads.detect_beads(tp_array)

                aprint(f"Done extracting PSF from time point: {i}/{len(time_points)} .")

            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {i} !")
                import traceback

                traceback.print_exc()

                if stop_at_exception:
                    raise error

            psfs.append(estimated_psf)

        psfs = np.stack(psfs).mean(axis=0)
        print(psfs.shape)
        np.save(dest_path + channel + ".npy", psfs)



import cupy

from dexp.utils.backends import Backend
from dexp.utils.backends._cupy.texture.texture import create_cuda_texture


def _warp_2d_cupy(image, vector_field, mode, block_size: int = 16):
    """

    Parameters
    ----------
    image
    vector_field
    mode
    block_size

    Returns
    -------

    """
    xp = Backend.get_xp_module()
    source = r"""
                extern "C"{
                __global__ void warp_2d(float* warped_image,
                                        cudaTextureObject_t input_image,
                                        cudaTextureObject_t vector_field,
                                        int width,
                                        int height)
                {
                    unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;
                    unsigned int y = blockIdx.y * blockDim.y + threadIdx.y;

                    if (x < width && y < height)
                    {
                        // coordinates in coord-normalised vector_field texture:
                        float u = float(x)/width;
                        float v = float(y)/height;

                        // Obtain linearly interpolated vector at (u,v):
                        float2 vector = tex2D<float2>(vector_field, u, v);

                        // Obtain the shifted coordinates of the source voxel:
                        // flip axis order to match numpy order
                        float sx = 0.5f + float(x) - vector.y;
                        float sy = 0.5f + float(y) - vector.x;

                        // Sample source image for voxel value:
                        float value = tex2D<float>(input_image, sx, sy);

                        //printf("(%f, %f)=%f\n", sx, sy, value);

                        // Store interpolated value:
                        warped_image[y * width + x] = value;

                        // TODO: supersampling would help in regions for which warping misses voxels in
                        // the source image.
                        // improve: adaptive supersampling would automatically use the vector field divergence to
                        // determine where to super sample and by how much.
                    }
                }
                }
                """

    if image.ndim != 2 or vector_field.ndim != 3:
        raise ValueError("image or vector field has wrong number of dimensions!")

    # set up textures:
    input_image_tex, input_image_cudarr = create_cuda_texture(
        image,
        num_channels=1,
        normalised_coords=False,
        sampling_mode="linear",
        address_mode=mode,
    )

    vector_field_tex, vector_field_cudarr = create_cuda_texture(
        vector_field, num_channels=2, normalised_coords=True, sampling_mode="linear", address_mode="clamp"
    )

    # Set up resulting image:
    warped_image = xp.empty(shape=image.shape, dtype=image.dtype)

    # get the kernel, which copies from texture memory
    warp_2d_kernel = cupy.RawKernel(source, "warp_2d")

    # launch kernel
    height, width = image.shape

    grid_x = (width + block_size - 1) // block_size
    grid_y = (height + block_size - 1) // block_size
    warp_2d_kernel(
        (grid_x, grid_y), (block_size,) * 2, (warped_image, input_image_tex, vector_field_tex, width, height)
    )

    del input_image_tex, input_image_cudarr, vector_field_tex, vector_field_cudarr

    return warped_image



from os import makedirs
from os.path import exists, join
from typing import Any, Sequence, Tuple

import imageio
from arbol.arbol import aprint, asection
from joblib import Parallel, delayed

from dexp.datasets import BaseDataset
from dexp.processing.color.projection import project_image
from dexp.utils.backends import Backend, BestBackend


def dataset_projection_rendering(
    input_dataset: BaseDataset,
    output_path: str = None,
    channels: Sequence[str] = None,
    slicing: Any = None,
    overwrite: bool = False,
    axis: int = 0,
    dir: int = -1,
    mode: str = "colormax",
    clim: Tuple[float, float] = None,
    attenuation: float = 0.05,
    gamma: float = 1.0,
    dlim: Tuple[float, float] = None,
    colormap: str = None,
    rgbgamma: float = 1.0,
    transparency: bool = False,
    legendsize: float = 1.0,
    legendscale: float = 1.0,
    legendtitle: str = "color-coded depth (voxels)",
    legendtitlecolor: Tuple[float, float, float, float] = (1, 1, 1, 1),
    legendposition: str = "bottom_left",
    legendalpha: float = 1.0,
    step: int = 1,
    workers: int = -1,
    workersbackend: str = "threading",
    devices: Tuple[int, ...] = (0,),
    stop_at_exception=True,
):
    for channel in channels:

        # Ensures that the output folder exists per channel:
        if len(channels) == 1:
            channel_output_path = output_path
        else:
            channel_output_path = output_path + f"_{channel}"

        makedirs(channel_output_path, exist_ok=True)

        with asection(f"Channel '{channel}' shape: {input_dataset.shape(channel)}:"):
            aprint(input_dataset.info(channel))

        array = input_dataset.get_array(channel, wrap_with_dask=True)

        if slicing:
            array = array[slicing]

        aprint(f"Rendering array of shape={array.shape} and dtype={array.dtype} for channel '{channel}'.")

        nbframes = array.shape[0]

        with asection("Rendering:"):

            def process(tp, _clim, device):
                try:
                    with asection(f"Rendering Frame     : {tp:05}"):

                        filename = join(channel_output_path, f"frame_{tp:05}.png")

                        if overwrite or not exists(filename):

                            with asection("Loading stack..."):
                                stack = array[tp].compute()

                            with BestBackend(device, exclusive=True, enable_unified_memory=True):
                                if _clim is not None:
                                    aprint(f"Using provided min and max for contrast limits: {_clim}")
                                    min_value, max_value = (float(strvalue) for strvalue in _clim.split(","))
                                    _clim = (min_value, max_value)

                                with asection(f"Projecting image of shape: {stack.shape} "):
                                    projection = project_image(
                                        stack,
                                        axis=axis,
                                        dir=dir,
                                        mode=mode,
                                        attenuation=attenuation,
                                        attenuation_min_density=0.002,
                                        attenuation_filtering=4,
                                        gamma=gamma,
                                        clim=_clim,
                                        cmap=colormap,
                                        dlim=dlim,
                                        rgb_gamma=rgbgamma,
                                        transparency=transparency,
                                        legend_size=legendsize,
                                        legend_scale=legendscale,
                                        legend_title=legendtitle,
                                        legend_title_color=legendtitlecolor,
                                        legend_position=legendposition,
                                        legend_alpha=legendalpha,
                                    )

                                with asection(f"Saving frame {tp} as: {filename}"):
                                    imageio.imwrite(filename, Backend.to_numpy(projection), compress_level=1)

                except Exception as error:
                    aprint(error)
                    aprint(f"Error occurred while processing time point {tp} !")
                    import traceback

                    traceback.print_exc()

                    if stop_at_exception:
                        raise error

            if workers == -1:
                workers = len(devices)
            aprint(f"Number of workers: {workers}")

            if workers > 1:
                Parallel(n_jobs=workers, backend=workersbackend)(
                    delayed(process)(tp, clim, devices[tp % len(devices)]) for tp in range(0, nbframes, step)
                )
            else:
                for tp in range(0, nbframes, step):
                    process(tp, clim, devices[0])



import click
from arbol.arbol import aprint, asection

from dexp.cli.parsing import _parse_channels, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.view import dataset_view
from dexp.datasets.operations.view_remote import dataset_view_remote


@click.command()
@click.argument("input_paths", nargs=-1)
@click.option("--channels", "-c", default=None, help="list of channels, all channels when ommited.")
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z).",
)
@click.option("--aspect", "-a", type=float, default=4, help="sets aspect ratio e.g. 4", show_default=True)
@click.option(
    "--rescale-time",
    "-rt",
    is_flag=True,
    help="Rescale time axis, useful when channels acquisition rate was not uniform.",
    show_default=True,
)
@click.option(
    "--clim",
    "-cl",
    type=str,
    default="0,512",
    help="Sets the contrast limits, i.e. -cl 0,1000 sets the contrast limits to [0,1000]",
    show_default=True,
)
@click.option(
    "--colormap",
    "-cm",
    type=str,
    default="viridis",
    help="sets colormap, e.g. viridis, gray, magma, plasma, inferno ",
    show_default=True,
)
@click.option(
    "--windowsize",
    "-ws",
    type=int,
    default=1536,
    help="Sets the napari window size. i.e. -ws 400 sets the window to 400x400",
    show_default=True,
)
@click.option(
    "--projectionsonly", "-po", is_flag=True, help="To view only the projections, if present.", show_default=True
)
@click.option("--volumeonly", "-vo", is_flag=True, help="To view only the volumetric data.", show_default=True)
def view(input_paths, channels, slicing, aspect, rescale_time, clim, colormap, windowsize, projectionsonly, volumeonly):
    """Views dataset using napari (napari.org)"""

    slicing = _parse_slicing(slicing)

    name = input_paths[0] + "..." if len(input_paths) > 1 else ""

    contrast_limits = tuple(float(v.strip()) for v in clim.split(","))

    if len(input_paths) == 1 and "http" in input_paths[0]:

        with asection(f"Viewing dataset at: {input_paths}, channels: {channels}, slicing: {slicing}, aspect:{aspect} "):
            dataset_view_remote(
                input_path=input_paths[0],
                name=name,
                aspect=aspect,
                channels=channels,
                contrast_limits=contrast_limits,
                colormap=colormap,
                slicing=slicing,
                windowsize=windowsize,
            )
            aprint("Done!")

    else:
        input_dataset, input_paths = glob_datasets(input_paths)
        channels = _parse_channels(input_dataset, channels)

        with asection(f"Viewing dataset at: {input_paths}, channels: {channels}, slicing: {slicing}, aspect:{aspect} "):
            dataset_view(
                input_dataset=input_dataset,
                name=name,
                aspect=aspect,
                channels=channels,
                contrast_limits=contrast_limits,
                colormap=colormap,
                slicing=slicing,
                windowsize=windowsize,
                projections_only=projectionsonly,
                volume_only=volumeonly,
                rescale_time=rescale_time,
            )
            input_dataset.close()
            aprint("Done!")



from typing import List, Optional, Tuple, Union

import numpy as np
from arbol import aprint
from numpy.typing import ArrayLike
from scipy.signal._signaltools import _centered

from dexp.utils import xpArray
from dexp.utils.backends import Backend


class Bead:
    def __init__(self, image: ArrayLike, coord: Optional[Tuple[int]], size: int):
        if coord is not None:
            self.slicing = self.get_slice(coord, size)
            self.data = image[self.slicing]
        else:
            self.slicing = None
            self.data = image
        self.data = Backend.current().to_numpy(self.data)
        self.data = self.data / np.linalg.norm(self.data)

    @property
    def shape(self) -> Tuple[int]:
        return self.data.shape

    @staticmethod
    def get_slice(coord: Tuple[int], size: int) -> List[slice]:
        slicing = tuple(slice(c - size // 2, c + size // 2 + 1) for c in coord)
        return slicing

    @staticmethod
    def from_data(array: ArrayLike) -> "Bead":
        return Bead(array, None, -1)

    def __array__(self, *args) -> ArrayLike:
        return self.data.reshape(-1)


class BeadsRemover:
    def __init__(self, peak_threshold: int, similarity_threshold: float, psf_size: int = 31, verbose: bool = False):
        """
        Helper object to detect beads, extract the PSF from it and remove it if necessary.

        Args:
            peak_threshold (int): Peak maxima threshold for object (beads) detection.
                Lower values detect more objects, but it might get more false positives.
            similarity_threshold (float): Threshold for beads selection given their cosine
                similarity with the estimated (median) PSF.
            psf_size (int, optional): PSF size (shape). Defaults to 31.
            verbose (bool, optional): Flag to display intermediate results. Defaults to False.
        """
        self.peak_threshold = peak_threshold
        self.similarity_threshold = similarity_threshold
        self.psf_size = psf_size
        self.beads: List[Bead] = []

        self.verbose = verbose

    def detect_beads(self, array: xpArray, estimated_psf: Optional[ArrayLike] = None) -> ArrayLike:
        from cucim.skimage.feature import peak_local_max

        backend = Backend.current()

        sp = backend.get_sp_module()
        xp = backend.get_xp_module()

        if estimated_psf is None:
            aprint("Finding beads with point detection.")
            kernel = xp.zeros((1, 5, 5))
            kernel[0, 2, 2] = 4
            kernel[0, 2, 0] = -1
            kernel[0, 0, 2] = -1
            kernel[0, 2, 4] = -1
            kernel[0, 4, 2] = -1
        else:
            aprint("Finding beads with estimated PSF.")
            kernel = estimated_psf.copy()
            dark_region = kernel < kernel.mean()
            bright_region = np.logical_not(dark_region)
            # equalize the values of both the regions to be 1.0
            kernel[dark_region] = -1.0 / dark_region.sum()
            kernel[bright_region] /= kernel[bright_region].sum()

            kernel = backend.to_backend(kernel.astype(xp.float32))

        # try detecting beads with given PSF or dot detection
        filtered = sp.ndimage.correlate(array.astype(xp.float32), weights=kernel)

        # small blur to center detection
        mean_kernel = xp.ones((5,) * array.ndim, dtype=xp.float32)
        mean_kernel /= mean_kernel.sum()
        filtered = sp.ndimage.correlate(filtered, weights=mean_kernel)

        # find local maxima in filtered image
        coordinates = peak_local_max(filtered, min_distance=self.psf_size, threshold_abs=500).get()

        # find local maxima in original image near the filtered maxima
        peaks = xp.zeros(array.shape, dtype=bool)
        peaks[tuple(coordinates.T)] = True
        peaks = sp.ndimage.binary_dilation(peaks, xp.ones((self.psf_size,) * array.ndim, dtype=bool))
        coordinates = peak_local_max(array, min_distance=self.psf_size, labels=peaks).get()

        if self.verbose:
            import napari

            viewer = napari.Viewer()
            viewer.add_image(
                filtered.get(), name="filtered", blending="additive", colormap="red", contrast_limits=(0, 5000)
            )
            viewer.add_image(
                array.get(), name="original", blending="additive", colormap="green", contrast_limits=(0, 5000)
            )
            peaks = xp.zeros(array.shape, dtype=bool)
            peaks[tuple(coordinates.T)] = True
            viewer.add_image(peaks.get(), name="peaks", blending="additive", colormap="gray", contrast_limits=(0, 1))
            napari.run()

        array = array.get()
        beads = []
        for coord in coordinates:
            bead = Bead(array, coord, self.psf_size)
            if np.all(np.equal(bead.shape, self.psf_size)):
                beads.append(bead)

        avg_bead = Bead.from_data(np.median(np.stack(tuple(beads)), axis=0).reshape((self.psf_size,) * array.ndim))

        # select regions that are similar to the median estimated bead
        selected_beads = []
        for bead in beads:
            angle = np.dot(bead, avg_bead)
            if angle > self.similarity_threshold:
                selected_beads.append(bead)
                array[bead.slicing] = 255

        aprint(f"Detected {len(selected_beads)} beads.")

        if self.verbose:
            import napari

            viewer = napari.Viewer(ndisplay=3)
            viewer.add_image(avg_bead.data, name="estimated PSF", interpolation="nearest")
            viewer.add_image(array, contrast_limits=(0, 500), name="marked beads")
            napari.run()

        self.beads = selected_beads
        return avg_bead.data

    def remove_beads(self, array: xpArray, fill_value: int = 0) -> xpArray:
        for bead in self.beads:
            array[bead.slicing] = fill_value
        return array


def remove_beads_by_threshold(
    image: xpArray, psf_size: Union[int, Tuple[int]] = 35, window: int = 7, k: float = 0.01
) -> xpArray:
    """
    This functions estimates beads by a local threshold and
    replace them with the minimum value of their respective region

    Parameters
    ----------
    image : xpArray
        Input image, validated with raw data
    psf_size : Union[int, Tuple[int]], optional
        Estimated PSF size for removal after detection, by default 35
    window : int, optional
        Local threshold window, by default 15
    k : float, optional
        Local threshold std deviation parameter, by default 0.05
        Reference: https://scikit-image.org/docs/dev/api/skimage.filters.html#skimage.filters.threshold_sauvola

    Returns
    -------
    xpArray
        Input image without beads
    """
    from cucim.skimage.filters import threshold_niblack

    sp = Backend.get_sp_module(image)
    xp = Backend.get_xp_module(image)

    shape = image.shape
    sampling = 4

    # downsample and remove noise from image
    small_image = sp.ndimage.zoom(image, zoom=1 / sampling, order=1)
    small_image = sp.ndimage.convolve(small_image, xp.ones((3,) * image.ndim))

    # apply threshold and upsample
    threshold = threshold_niblack(small_image, window, k)
    threshold = sp.ndimage.zoom(threshold, zoom=sampling, order=1)
    threshold = _centered(threshold, shape)

    # detect beads
    detection = image > threshold
    detection = sp.ndimage.minimum_filter(detection, 2)  # erosion

    # dilate and fill regions
    detection = sp.ndimage.maximum_filter(detection, psf_size)  # dilation

    # replacing mask with minimum
    min_image = sp.ndimage.minimum_filter(image, psf_size)
    output = image.copy()
    output[detection] = min_image[detection]

    return output



from dexp.utils.backends import Backend


def mean_squared_error(image_a, image_b):
    # Backend:
    xp = Backend.get_xp_module(image_a)
    return xp.mean(xp.square(image_a - image_b))


def mean_absolute_error(image_a, image_b):
    # Backend:
    xp = Backend.get_xp_module(image_a)
    return xp.mean(xp.absolute(image_a - image_b))


def lhalf_error(image_a, image_b):
    # Backend:
    xp = Backend.get_xp_module(image_a)
    return xp.square(xp.mean(xp.sqrt(xp.absolute(image_a - image_b))))


def psnr(image_true, image_test):
    """
    Compute the peak signal to noise ratio (PSNR) for a [0,1] normalised image.

    Parameters
    ----------
    image_true : ndarray
        Ground-truth image, same shape as im_test.
    image_test : ndarray
        Test image.
    data_range : int, optional
        The data range of the input image (distance between minimum and
        maximum possible values).  By default, this is estimated from the image
        data-type.

    Returns
    -------
    psnr : float
        The PSNR metric.


    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio

    """
    # Backend:
    xp = Backend.get_xp_module()

    # Move to backend:
    image_true = Backend.to_backend(image_true, dtype=xp.float32)
    image_test = Backend.to_backend(image_test, dtype=xp.float32)

    err = mean_squared_error(image_true, image_test)
    return 10 * xp.log10(1 / err)


def ssim(image_a, image_b, win_size=None):
    """
    Compute the mean structural similarity index between two images.

    Parameters
    ----------
    image_a, image_b : ndarray
        Images. Any dimensionality with same shape.
    win_size : int or None, optional
        The side-length of the sliding window used in comparison. Must be an
        odd value. If `gaussian_weights` is True, this is ignored and the
        window size will depend on `sigma`.


    Returns
    -------
    mssim : float
        The mean structural similarity index over the image.



    References
    ----------
    .. [1] Wang, Z., Bovik, A. C., Sheikh, H. R., & Simoncelli, E. P.
       (2004). Image quality assessment: From error visibility to
       structural similarity. IEEE Transactions on Image Processing,
       13, 600-612.
       https://ece.uwaterloo.ca/~z70wang/publications/ssim.pdf,
       :DOI:`10.1109/TIP.2003.819861`

    .. [2] Avanaki, A. N. (2009). Exact global histogram specification
       optimized for structural similarity. Optical Review, 16, 613-621.
       :arxiv:`0901.0065`
       :DOI:`10.1007/s10043-009-0119-z`

    """

    # Backend:
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    # Move to backend:
    image_a = Backend.to_backend(image_a, dtype=xp.float32)
    image_b = Backend.to_backend(image_b, dtype=xp.float32)

    if win_size is None:
        win_size = 7  # backwards compatibility

    if not (win_size % 2 == 1):
        raise ValueError("Window size must be odd.")

    # Constants:
    K1 = 0.01
    K2 = 0.03

    filter_func = sp.ndimage.uniform_filter
    filter_args = {"size": win_size}

    # filter has already normalized by NP
    cov_norm = 1.0  # population covariance to match Wang et. al. 2004

    # compute (weighted) means
    ux = filter_func(image_a, **filter_args)
    uy = filter_func(image_b, **filter_args)

    # compute (weighted) variances and covariances
    uxx = filter_func(image_a * image_a, **filter_args)
    uyy = filter_func(image_b * image_b, **filter_args)
    uxy = filter_func(image_a * image_b, **filter_args)
    vx = cov_norm * (uxx - ux * ux)
    vy = cov_norm * (uyy - uy * uy)
    vxy = cov_norm * (uxy - ux * uy)

    R = 1
    C1 = (K1 * R) ** 2
    C2 = (K2 * R) ** 2

    A1, A2, B1, B2 = (2 * ux * uy + C1, 2 * vxy + C2, ux ** 2 + uy ** 2 + C1, vx + vy + C2)
    D = B1 * B2
    S = (A1 * A2) / D

    # to avoid edge effects will ignore filter radius strip around edges
    pad = (win_size - 1) // 2

    # compute (weighted) mean of ssim
    mssim = _crop(S, pad).mean()

    return mssim


def _crop(ar, crop_width, copy=False, order="K"):
    """Crop array `ar` by `crop_width` along each dimension.

    Parameters
    ----------
    ar : array-like of rank N
        Input array.
    crop_width : {sequence, int}
        Number of values to remove from the edges of each axis.
        ``((before_1, after_1),`` ... ``(before_N, after_N))`` specifies
        unique crop widths at the start and end of each axis.
        ``((before, after),) or (before, after)`` specifies
        a fixed start and end crop for every axis.
        ``(n,)`` or ``n`` for integer ``n`` is a shortcut for
        before = after = ``n`` for all axes.
    copy : bool, optional
        If `True`, ensure the returned array is a contiguous copy. Normally,
        a crop operation will return a discontiguous view of the underlying
        input array.
    order : {'C', 'F', 'A', 'K'}, optional
        If ``copy==True``, control the memory layout of the copy. See
        ``np.copy``.

    Returns
    -------
    cropped : array
        The cropped array. If ``copy=False`` (default), this is a sliced
        view of the input array.
    """

    # Backend:
    xp = Backend.get_xp_module()

    ar = xp.array(ar, copy=False)

    if isinstance(crop_width, int):
        crops = [[crop_width, crop_width]] * ar.ndim
    elif isinstance(crop_width[0], int):
        if len(crop_width) == 1:
            crops = [[crop_width[0], crop_width[0]]] * ar.ndim
        elif len(crop_width) == 2:
            crops = [crop_width] * ar.ndim
        else:
            raise ValueError(
                f"crop_width has an invalid length: {len(crop_width)}\n"
                "crop_width should be a sequence of N pairs, "
                "a single pair, or a single integer"
            )
    elif len(crop_width) == 1:
        crops = [crop_width[0]] * ar.ndim
    elif len(crop_width) == ar.ndim:
        crops = crop_width
    else:
        raise ValueError(
            f"crop_width has an invalid length: {len(crop_width)}\n"
            "crop_width should be a sequence of N pairs, "
            "a single pair, or a single integer"
        )

    slices = tuple(slice(a, ar.shape[i] - b) for i, (a, b) in enumerate(crops))
    if copy:
        cropped = xp.array(ar[slices], order=order, copy=True)
    else:
        cropped = ar[slices]
    return cropped



from typing import Dict, Sequence

import numpy as np
from ome_zarr.format import CurrentFormat


def default_omero_metadata(name: str, channels: Sequence[str], dtype: np.dtype) -> Dict:
    val_max = 1.0 if np.issubdtype(dtype, np.floating) else np.iinfo(dtype).max

    return {
        "id": 1,
        "name": name,
        "version": CurrentFormat().version,
        "channels": [
            {
                "active": True,
                "coefficient": 1,
                "color": "FFFFFF",
                "family": "linear",
                "inverted": False,
                "label": ch,
                "window": {"start": 0, "end": 1500, "min": 0, "max": val_max},  # guessing
            }
            for ch in channels
        ],
    }



import inspect
from functools import wraps
from typing import Any, Callable, Dict, Iterable, List, Tuple

import numpy as np
import pytest

from dexp.utils.backends import Backend, CupyBackend, NumpyBackend


def _maybe_to_backend(backend: Backend, obj: Any) -> Any:

    if isinstance(obj, np.ndarray):
        # this must be first because arrays are iterables
        return backend.to_backend(obj)

    if isinstance(obj, Dict):
        dispatch_data_to_backend(backend, [], obj)
        return obj

    elif isinstance(obj, (List, Tuple)):
        obj = list(obj)
        dispatch_data_to_backend(backend, obj, {})
        return obj

    else:
        return obj


def dispatch_data_to_backend(backend: Backend, args: Iterable, kwargs: Dict) -> None:
    """Function to move arrays to backend."""
    for i, v in enumerate(args):
        args[i] = _maybe_to_backend(backend, v)

    for k, v in kwargs.items():
        kwargs[k] = _maybe_to_backend(backend, v)


def _add_cuda_signature(func: Callable) -> Callable:
    """Adds `cuda` argument to the given function."""
    # inspect parameters
    sig = inspect.signature(func)
    params = list(sig.parameters.values())

    # add new parameter
    params.insert(0, inspect.Parameter("cuda", kind=inspect.Parameter.POSITIONAL_OR_KEYWORD))

    # overwrite old signature
    func.__signature__ = sig.replace(parameters=params)

    # add cuda parameter to function
    parametrizer = pytest.mark.parametrize("cuda", [False, True], ids=["cpu", "gpu"])
    func = parametrizer(func)
    return func


def execute_both_backends(func: Callable) -> Callable:
    """Helper function to execute both backends in a single function call."""

    @wraps(func)
    def wrapper(cuda: bool, *args, **kwargs):
        args = list(args)
        if cuda:
            try:
                with CupyBackend() as backend:
                    dispatch_data_to_backend(backend, args, kwargs)
                    func(*args, **kwargs)
            except ModuleNotFoundError:
                pytest.skip(f"Cupy not found. Skipping {func.__name__} gpu test.")
        else:
            with NumpyBackend() as backend:
                # it assumes that is a numpy array by default
                func(*args, **kwargs)

    return _add_cuda_signature(wrapper)



from typing import Callable, Sequence, Tuple

import dask
from arbol import aprint, asection
from dask.distributed import Client
from dask_cuda import LocalCUDACluster
from toolz import curry

from dexp.datasets import BaseDataset, ZDataset
from dexp.datasets.stack_iterator import StackIterator
from dexp.processing.denoising import calibrate_denoise_butterworth, denoise_butterworth
from dexp.processing.utils.scatter_gather_i2i import scatter_gather_i2i
from dexp.utils.backends import CupyBackend


@curry
def _process(
    stacks: StackIterator,
    out_dataset: ZDataset,
    channel: str,
    time_point: int,
    scatter_gather: Callable,
) -> None:

    with CupyBackend() as bkd:
        with asection(f"Calibrating and denoising channel {channel} time point {time_point}"):
            stack = bkd.to_backend(stacks[time_point])
            _, best_params = calibrate_denoise_butterworth(stack)
            denoise_fun = curry(denoise_butterworth, **best_params)

            with asection("Denoising"):
                denoised = scatter_gather(function=denoise_fun, image=stack)

            out_dataset.write_stack(channel, time_point, bkd.to_numpy(denoised))


def dataset_denoise(
    input_dataset: BaseDataset,
    output_dataset: ZDataset,
    channels: Sequence[str],
    tilesize: Tuple[int],
    devices: Sequence[int],
):
    # NOTE:
    #   we might want in the future to smooth the parameters estimation over neighboring time points.
    lazy_computations = []

    for ch in channels:
        stacks = input_dataset[ch]
        output_dataset.add_channel(ch, stacks.shape, dtype=input_dataset.dtype(ch))

        # Create processing function with default parameters
        process = dask.delayed(
            _process(
                stacks=stacks,
                out_dataset=output_dataset,
                channel=ch,
                scatter_gather=curry(scatter_gather_i2i, tiles=tilesize, margins=32),
            )
        )  # using 32 because Jordao assumed it's good enough and 320 (default tile) + 64 = 384
        # has a nice prime factorization, speeding up fft computation

        # Stores functions to be computed
        lazy_computations += [process(time_point=t) for t in range(len(stacks))]

    cluster = LocalCUDACluster(CUDA_VISIBLE_DEVICES=devices)
    client = Client(cluster)
    aprint("Dask client", client)

    # Compute everything
    dask.compute(*lazy_computations)

    output_dataset.check_integrity()



import click

from dexp.cli.defaults import DEFAULT_WORKERS_BACKEND
from dexp.video.overlay import add_overlays_image_sequence


@click.command()
@click.argument("input_path", type=str)
@click.option("--output_path", "-o", type=str, default=None, help="Output folder for overlayed frames.")
@click.option("--scalebar/--no-scalebar", "-sb/-nsb", default=True, help="True to insert scale bar.", show_default=True)
@click.option(
    "--barlength", "-bl", type=float, default=1, help="Length of scale bar in the provided unit.", show_default=True
)
@click.option(
    "--barscale",
    "-bs",
    type=float,
    default=1,
    help="Conversion factor from pixels to units -- what is the side length of a pixel/voxel in units.",
    show_default=True,
)
@click.option("--barheight", "-bh", type=int, default=4, help="Height of th scale bar in pixels", show_default=True)
@click.option(
    "--barpos",
    "-bp",
    type=str,
    default="bottom_right",
    help="Positions of the scale bar in pixels in natural order: (x, y)."
    " Can also be a string: bottom_left, bottom_right, top_left, top_right.",
    show_default=True,
)
@click.option("--barunit", "-bu", type=str, default="μm", help="Scale bar unit name.", show_default=True)
@click.option(
    "--timestamp/--no-timestamp", "-ts/-nts", default=True, help="True to insert time stamp.", show_default=True
)
@click.option("--timestart", "-ts", type=float, default=0, help="Start time for time stamp", show_default=True)
@click.option(
    "--timeinterval",
    "-ti",
    type=float,
    default=1,
    help="Time interval inn units of time between consecutive images.",
    show_default=True,
)
@click.option(
    "--timepos",
    "-tp",
    type=str,
    default="top_right",
    help="Positions of the time stamp in pixels in natural order: (x, y)."
    " Can also be a string: bottom_left, bottom_right, top_left, top_right.",
    show_default=True,
)
@click.option("--timeunit", "-tu", type=str, default="s", help="Time stamp time unit name.", show_default=True)
@click.option(
    "--margin",
    "-mg",
    type=float,
    default=1,
    help="Margin around bar expressed in units relative to the text height",
    show_default=True,
)
@click.option(
    "--color",
    "-c",
    type=str,
    default=None,
    help="Color of the bar and text as tuple of 4 values: (R, G, B, A)",
    show_default=True,
)
@click.option(
    "--numberformat",
    "-nf",
    type=str,
    default="{:.1f}",
    help="Format string to represent the start and end values.",
    show_default=True,
)
@click.option("--fontname", "-fn", type=str, default="Helvetica", help="Font name.", show_default=True)
@click.option("--fontsize", "-fs", type=int, default=32, help="Font size in pixels.", show_default=True)
@click.option(
    "--mode",
    "-md",
    type=str,
    default="max",
    help="Blending modes. Either one for all images, or one per image in the form of a sequence."
    " Blending modes are: mean, add, satadd, max, alpha.",
    show_default=True,
)
@click.option("--overwrite", "-w", is_flag=True, help="Force overwrite of output images.", show_default=True)
@click.option(
    "--workers",
    "-k",
    type=int,
    default=-1,
    help="Number of worker threads to spawn, set to -1 for maximum number of workers",
    show_default=True,
)  #
@click.option(
    "--workersbackend",
    "-wkb",
    type=str,
    default=DEFAULT_WORKERS_BACKEND,
    help="What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ",
    show_default=True,
)  #
@click.option("--device", "-d", type=int, default=0, help="Sets the CUDA devices id, e.g. 0,1,2", show_default=True)  #
def overlay(
    input_path: str,
    output_path: str,
    scalebar,
    barlength,
    barscale,
    barheight,
    barpos,
    barunit,
    timestamp,
    timestart,
    timeinterval,
    timepos,
    timeunit,
    margin,
    color,
    numberformat,
    fontname,
    fontsize,
    mode,
    overwrite,
    workers,
    workersbackend,
    device,
):
    """Adds scale bars and timestamps to videos."""

    # Default output path:
    if output_path is None:
        output_path = input_path + "_overlay"
    elif output_path.startswith("_"):
        output_path = input_path + output_path

    # Parse bar position:
    if "," in barpos:
        barpos = tuple(float(v) for v in barpos.split(","))

    # Parse time position:
    if "," in timepos:
        timepos = tuple(float(v) for v in timepos.split(","))

    # Parse color:
    if color is not None and "," in color:
        color = tuple(float(v) for v in color.split(","))

    add_overlays_image_sequence(
        input_path=input_path,
        output_path=output_path,
        scale_bar=scalebar,
        scale_bar_length_in_unit=barlength,
        scale_bar_pixel_scale=barscale,
        scale_bar_bar_height=barheight,
        scale_bar_translation=barpos,
        scale_bar_unit=barunit,
        time_stamp=timestamp,
        time_stamp_start_time=timestart,
        time_stamp_time_interval=timeinterval,
        time_stamp_translation=timepos,
        time_stamp_unit=timeunit,
        margin=margin,
        color=color,
        number_format=numberformat,
        font_name=fontname,
        font_size=fontsize,
        mode=mode,
        overwrite=overwrite,
        workers=workers,
        workersbackend=workersbackend,
        device=device,
    )



import click
from arbol.arbol import aprint

from dexp.cli.video_commands.blend import blend
from dexp.cli.video_commands.mp4 import mp4
from dexp.cli.video_commands.overlay import overlay
from dexp.cli.video_commands.resize import resize
from dexp.cli.video_commands.stack import stack

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group()
def cli():
    aprint("__________________________________________")
    aprint("  DEXP -- Data EXploration & Processing   ")
    aprint("  Video editing dexp_commands                  ")
    aprint("  Royer lab                               ")
    aprint("__________________________________________")
    aprint("")
    aprint(
        "Note: videos are folders of images (typically PNGs), final step is to convert such image sequences into MP4s"
    )
    # input("Press Enter to continue...")


cli.add_command(overlay)
cli.add_command(blend)
cli.add_command(stack)
cli.add_command(resize)
cli.add_command(mp4)



from typing import Callable, Tuple

import numpy as np

from dexp.processing.utils.element_wise_affine import element_wise_affine
from dexp.utils import xpArray
from dexp.utils.backends import Backend, NumpyBackend


class Normalise:
    def __init__(
        self,
        image: xpArray,
        low: float = 0.0,
        high: float = 1.0,
        minmax: Tuple[float, float] = None,
        quantile: float = 0,
        clip: bool = False,
        do_normalise: bool = True,
        in_place: bool = True,
        dtype=None,
    ) -> Tuple[Callable, Callable]:
        """Returns a pair of functions: the first normalises the given image to the range [low, high],
        the second denormalises back to the original range (and dtype).
        Usefull when determining the normalisation parameters and doing the actual normalisation must be decoupled,
        for example when splitting an image into chunks and sensing computation to the GPU.


        Parameters
        ----------
        image : image to normalise
        low, high : normalisation range
        minmax : min and max values of the image if already known.
        quantile : if quantile>0 then quantile normalisation is used to find the min and max values.
            Value must be within [0,1]
        clip : clip after normalisation/denormalisation
        do_normalise : If False, the returned functions are pass-through identity functions
            -- usefull to turn on and off normalisation while still using the functions themselves.
        in_place : In-place computation is allowed and inputs may be modified
        dtype : dtype to normalise to.

        Returns
        -------

        """
        xp = Backend.get_xp_module()

        if isinstance(Backend.current(), NumpyBackend):
            dtype = np.float32

        elif dtype is None:
            if image.itemsize <= 2:
                dtype = xp.float16
            elif image.itemsize == 4 or image.itemsize == 8:
                dtype = xp.float32
            else:
                raise ValueError(f"Failed to converted type {image.dtype}")

        self.dtype = dtype
        self.original_dtype = image.dtype
        self.clip = clip
        self.low = low
        self.high = high
        self.min_value = None
        self.max_value = None
        self.in_place = in_place
        self.do_normalise = do_normalise
        self.norm_alpha = None
        self.norm_beta = None
        self.denorm_alpha = None
        self.denorm_beta = None

        if not self.do_normalise:
            return

        if minmax is None:
            min_value, max_value = 0, 0
            if quantile > 0:
                min_value = xp.quantile(image, q=quantile)
                max_value = xp.quantile(image, q=1 - quantile)

            # if we did not use quantiles or we get some weird result, let's fallback to standard min max:
            if min_value >= max_value:
                min_value = xp.min(image)
                max_value = xp.max(image)

        else:
            min_value, max_value = minmax

        self.min_value = min_value
        self.max_value = max_value

        # Normalise:
        norm_denom = max_value - min_value
        if norm_denom == 0:
            norm_denom = 1

        norm_alpha = (high - low) / norm_denom
        norm_beta = low - norm_alpha * min_value

        # Ensure correct type:
        self.norm_alpha = xp.asarray(norm_alpha, dtype=dtype)
        self.norm_beta = xp.asarray(norm_beta, dtype=dtype)

        # Denormalise:
        denorm_alpha = (max_value - min_value) / (high - low)
        denorm_beta = min_value - denorm_alpha * low

        # Ensure correct type:
        self.denorm_alpha = xp.asarray(denorm_alpha, dtype=dtype)
        self.denorm_beta = xp.asarray(denorm_beta, dtype=dtype)

    def _transform(
        self, array: xpArray, alpha: float, beta: float, low: float, high: float, dtype: np.dtype
    ) -> xpArray:

        if self.do_normalise:
            array = Backend.to_backend(array, dtype=self.dtype)
            out = array if self.in_place else None
            array = element_wise_affine(array, alpha, beta, out=out)
            if self.clip:
                array = array.clip(low, high, out=array)
        return array.astype(dtype, copy=False)

    def forward(self, array: xpArray) -> xpArray:
        return self._transform(array, self.norm_alpha, self.norm_beta, self.low, self.high, self.dtype)

    def backward(self, array: xpArray) -> xpArray:
        return self._transform(
            array, self.denorm_alpha, self.denorm_beta, self.min_value, self.max_value, self.original_dtype
        )



import dask
import numpy as np
from arbol.arbol import aprint, asection
from toolz import curry

from dexp.datasets import ZDataset
from dexp.processing.multiview_lightsheet.fusion.mvsols import msols_fuse_1C2L
from dexp.processing.multiview_lightsheet.fusion.simview import SimViewFusion
from dexp.processing.registration.model.model_io import (
    model_list_from_file,
    model_list_to_file,
)
from dexp.utils.backends import Backend, BestBackend, NumpyBackend
from dexp.utils.slicing import slice_from_shape


def dataset_fuse(
    dataset,
    output_path,
    channels,
    slicing,
    store,
    compression,
    compression_level,
    overwrite,
    microscope,
    equalise,
    equalise_mode,
    zero_level,
    clip_too_high,
    fusion,
    fusion_bias_strength_i,
    fusion_bias_strength_d,
    dehaze_size,
    dark_denoise_threshold,
    z_pad_apodise,
    loadreg,
    model_list_filename,
    warpreg_num_iterations,
    min_confidence,
    max_change,
    registration_edge_filter,
    maxproj,
    huge_dataset,
    devices,
    check,
    pad,
    white_top_hat_size,
    white_top_hat_sampling,
    remove_beads,
    stop_at_exception=True,
):

    views = {channel.split("-")[-1]: dataset.get_array(channel, per_z_slice=False) for channel in channels}

    with asection("Views:"):
        for channel, view in views.items():
            aprint(f"View: {channel} of shape: {view.shape} and dtype: {view.dtype}")

    key = list(views.keys())[0]
    dtype = views[key].dtype
    in_nb_time_points = views[key].shape[0]
    aprint(f"Slicing with: {slicing}")
    _, volume_slicing, time_points = slice_from_shape(views[key].shape, slicing)

    if microscope == "simview":
        views = SimViewFusion.validate_views(views)

    # load registration models:
    with NumpyBackend():
        if loadreg:
            aprint(f"Loading registration shifts from existing file! ({model_list_filename})")
            models = model_list_from_file(model_list_filename)
            if len(models) == 1:
                models = [models[0] for _ in range(in_nb_time_points)]
            elif len(models) != in_nb_time_points:
                raise ValueError(
                    f"Number of registration models provided ({len(models)})"
                    f"differs from number of input time points ({in_nb_time_points})"
                )
        else:
            models = [None] * len(time_points)

    @curry
    def process(i, params, device_id=0):
        equalisation_ratios_reference, model, dest_dataset = params
        tp = time_points[i]
        try:
            with asection(f"Fusing time point for time point {i}/{len(time_points)}"):
                with asection(f"Loading channels {channels}"):
                    views_tp = {k: np.asarray(view[tp][volume_slicing]) for k, view in views.items()}

                with BestBackend(exclusive=True, enable_unified_memory=True, device_id=device_id):
                    if models[i] is not None:
                        model = models[i]
                    # otherwise it could be a None model or from the first iteration if equalisation mode was 'first'
                    if microscope == "simview":
                        fuse_obj = SimViewFusion(
                            registration_model=model,
                            equalise=equalise,
                            equalisation_ratios=equalisation_ratios_reference,
                            zero_level=zero_level,
                            clip_too_high=clip_too_high,
                            fusion=fusion,
                            fusion_bias_exponent=2,
                            fusion_bias_strength_i=fusion_bias_strength_i,
                            fusion_bias_strength_d=fusion_bias_strength_d,
                            dehaze_before_fusion=True,
                            dehaze_size=dehaze_size,
                            dehaze_correct_max_level=True,
                            dark_denoise_threshold=dark_denoise_threshold,
                            dark_denoise_size=9,
                            white_top_hat_size=white_top_hat_size,
                            white_top_hat_sampling=white_top_hat_sampling,
                            remove_beads=remove_beads,
                            butterworth_filter_cutoff=1,
                            flip_camera1=True,
                            pad=pad,
                        )

                        tp_array = fuse_obj(**views_tp)
                        new_equalisation_ratios = fuse_obj._equalisation_ratios

                    elif microscope == "mvsols":
                        metadata = dataset.get_metadata()
                        angle = metadata["angle"]
                        dz = metadata["dz"]
                        res = metadata["res"]
                        illumination_correction_sigma = metadata["ic_sigma"] if "ic_sigma" in metadata else None

                        tp_array, model, new_equalisation_ratios = msols_fuse_1C2L(
                            *list(views_tp.values()),
                            z_pad=z_pad_apodise[0],
                            z_apodise=z_pad_apodise[1],
                            registration_num_iterations=warpreg_num_iterations,
                            registration_force_model=loadreg,
                            registration_model=model,
                            registration_min_confidence=min_confidence,
                            registration_max_change=max_change,
                            registration_edge_filter=registration_edge_filter,
                            equalise=equalise,
                            equalisation_ratios=equalisation_ratios_reference,
                            zero_level=zero_level,
                            clip_too_high=clip_too_high,
                            fusion=fusion,
                            fusion_bias_exponent=2,
                            fusion_bias_strength_x=fusion_bias_strength_i,
                            dehaze_size=dehaze_size,
                            dark_denoise_threshold=dark_denoise_threshold,
                            angle=angle,
                            dx=res,
                            dz=dz,
                            illumination_correction_sigma=illumination_correction_sigma,
                            registration_mode="projection" if maxproj else "full",
                            huge_dataset_mode=huge_dataset,
                        )
                    else:
                        raise NotImplementedError

                    with asection("Moving array from backend to numpy."):
                        tp_array = Backend.to_numpy(tp_array, dtype=dtype, force_copy=False)

                    del views_tp
                    Backend.current().clear_memory_pool()

                    aprint(f"Last equalisation ratios: {new_equalisation_ratios}")

                    # moving to numpy to allow pickling
                    if model is not None:
                        model = model.to_numpy()
                    new_equalisation_ratios = [
                        None if v is None else Backend.to_numpy(v) for v in new_equalisation_ratios
                    ]

                with asection(f"Saving fused stack for time point {i}, shape:{tp_array.shape}, dtype:{tp_array.dtype}"):

                    if i == 0:
                        # We allocate last minute once we know the shape... because we don't always know
                        # the shape in advance!!!
                        mode = "w" + ("" if overwrite else "-")
                        dest_dataset = ZDataset(output_path, mode, store, parent=dataset)
                        dest_dataset.add_channel(
                            "fused",
                            shape=(len(time_points),) + tp_array.shape,
                            dtype=tp_array.dtype,
                            codec=compression,
                            clevel=compression_level,
                        )

                    dest_dataset.write_stack(channel="fused", time_point=i, stack_array=tp_array)

                aprint(f"Done processing time point: {i}/{len(time_points)} .")

        except Exception as error:
            aprint(error)
            aprint(f"Error occurred while processing time point {i} !")
            import traceback

            traceback.print_exc()

            if stop_at_exception:
                raise error

        return new_equalisation_ratios, model, dest_dataset

    if len(devices) == 1:
        process = process(device_id=devices[0])
    else:
        from dask.distributed import Client
        from dask_cuda import LocalCUDACluster

        process = dask.delayed(process)

        cluster = LocalCUDACluster(CUDA_VISIBLE_DEVICES=devices)
        client = Client(cluster)

        aprint("Dask Client", client)

    # the parameters are (equalisation rations, registration model, dest. dataset)
    if microscope == "simview":
        init_params = ([None, None, None], None, None)
    elif microscope == "mvsols":
        init_params = ([None], None, None)
    else:
        raise NotImplementedError

    # it creates the output dataset from the first time point output shape
    params = process(0, init_params)
    if len(devices) > 1:
        params = params.persist()

    if equalise_mode == "all":
        params = init_params[:2] + (params[2],)

    lazy_computations = []  # it is only lazy if len(devices) > 1
    for i in range(1, len(time_points)):
        lazy_computations.append(process(i, params))

    if len(devices) > 1:
        if equalise_mode == "all":
            first_model, dest_dataset = params[1:]  # not a dask operation
        else:
            first_model, dest_dataset = params.compute()[1:]
        models = [first_model] + [output[1] for output in dask.compute(*lazy_computations)]
    else:
        models = [params[1]] + [output[1] for output in lazy_computations]
        dest_dataset = params[2]

    if not loadreg and models[0] is not None:
        model_list_to_file(model_list_filename, models)

    aprint(dest_dataset.info())
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()
    if len(devices) > 1:
        client.close()



import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import DEFAULT_CLEVEL, DEFAULT_CODEC, DEFAULT_STORE
from dexp.cli.parsing import _get_output_path, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.isonet import dataset_isonet


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option("--dxy", "-dxy", type=float, required=True)
@click.option("--dz", "-dz", type=float, required=True)
@click.option("--binning", "-b", type=int, default=1)  # TODO help
@click.option("--sharpening", "-sh", is_flag=True, help="Dehazes the image with a low-pass rejection filter.")
@click.option("--channel", "-ch", default=None, help="dataset channel name")
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)  #
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option(
    "--codec",
    "-z",
    default=DEFAULT_CODEC,
    help="compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ",
    show_default=True,
)  #
@click.option("--clevel", "-l", type=int, default=DEFAULT_CLEVEL, help="Compression level", show_default=True)
@click.option(
    "--overwrite", "-w", is_flag=True, help="to force overwrite of target", show_default=True
)  # , help='dataset slice'
@click.option(
    "--context", "-c", default="default", help="IsoNet context name", show_default=True
)  # , help='dataset slice'
@click.option(
    "--mode", "-m", default="pta", help="mode: 'pta' -> prepare, train and apply, 'a' just apply  ", show_default=True
)  # , help='dataset slice'
@click.option(
    "--max_epochs", "-e", type=int, default="50", help="to force overwrite of target", show_default=True
)  # , help='dataset slice'
@click.option("--check", "-ck", default=True, help="Checking integrity of written file.", show_default=True)  #
def isonet(
    input_paths,
    output_path,
    dxy,
    dz,
    sharpening,
    binning,
    channel,
    slicing,
    store,
    codec,
    clevel,
    overwrite,
    context,
    mode,
    max_epochs,
    check,
):
    """Recovers isotropic resolution using the ISONET approach (Weigert et al.)"""

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_isonet")
    slicing = _parse_slicing(slicing)

    with asection(f"Applying Isonet to: {input_paths}, saving result to: {output_path}, slicing: {slicing} "):
        dataset_isonet(
            input_dataset,
            output_path,
            dxy=dxy,
            dz=dz,
            channel=channel,
            slicing=slicing,
            store=store,
            compression=codec,
            compression_level=clevel,
            overwrite=overwrite,
            context=context,
            mode=mode,
            max_epochs=max_epochs,
            check=check,
            sharpening=sharpening,
            binning=binning,
            training_tp_index=None,
        )

        input_dataset.close()
        aprint("Done!")



import pytest

from dexp.datasets.synthetic_datasets import (
    binary_blobs,
    generate_fusion_test_data,
    generate_nuclei_background_data,
)


@pytest.fixture
def dexp_binary_blobs(request):
    return binary_blobs(request.param)


@pytest.fixture
def dexp_nuclei_background_data(request):
    return generate_nuclei_background_data(**request.param)


@pytest.fixture
def dexp_fusion_test_data(request):
    return generate_fusion_test_data(**request.param)


def pytest_addoption(parser):
    """pytest command line parser"""
    parser.addoption("--display", action="store", type=bool, default=False)


@pytest.fixture()
def display_test(pytestconfig):
    """display test fixture from pytest parser"""
    return pytestconfig.getoption("display")



from typing import Union

from dexp.processing.registration.model.pairwise_registration_model import (
    PairwiseRegistrationModel,
)
from dexp.processing.registration.model.sequence_registration_model import (
    SequenceRegistrationModel,
)

RegistrationModel = Union[SequenceRegistrationModel, PairwiseRegistrationModel]



from typing import Optional

import numexpr

from dexp.utils import xpArray
from dexp.utils.backends import Backend, CupyBackend


def element_wise_affine(
    array: xpArray, alpha: float, beta: float, sum_first: bool = False, out: Optional[xpArray] = None
) -> xpArray:

    """
    Applies the affine function: alpha*x + beta to every value x of a given array.
    If sum_first is True, then alpha*(x + beta) is computed instead.

    Parameters
    ----------
    array : array to apply function to
    alpha : 'scale'
    beta : 'offset'
    sum_first : If True then alpha*(x + beta) is computed instead.
    out : if specified, element_wise_affine _might_ use the specified array as output for in-place operation.
        this is not garanteed, so you should always use the pattern:
        result = element_wise_affine(backend, a, array, alpha, beta, out=result).


    Returns
    -------
    Array: alpha*array + beta (or alpha*array + beta if sum_first is True)

    """

    array = Backend.to_backend(array)

    if isinstance(Backend.current(), CupyBackend):
        import cupy

        if sum_first:

            @cupy.fuse()
            def affine_function(_array, _alpha, _beta):
                return _alpha * (_array + _beta)

        else:

            @cupy.fuse()
            def affine_function(_array, _alpha, _beta):
                return _alpha * _array + _beta

        return affine_function(array, alpha, beta)

    else:
        if sum_first:
            return numexpr.evaluate("alpha*(array+beta)", casting="same_kind", out=out)
        else:
            return numexpr.evaluate("alpha*array+beta", casting="same_kind", out=out)



import os
from os.path import exists

import click
from arbol.arbol import aprint, asection


@click.command()
@click.argument("input_path")
@click.option("--prefix", "-p", type=str, default="frame_", help="Prefix for image filenames")
@click.option("--leading", "-l", type=int, default=5, help="Number of leading zeros for frame indices.")
@click.option("--extension", "-e", type=str, default="png", help="Extension for image filenames")
@click.option("--output_path", "-o", type=str, default=None, help="Output file path for movie")
@click.option(
    "--framerate", "-fps", type=int, default=30, help="Sets the framerate in frames per second", show_default=True
)
@click.option(
    "--codec",
    "-c",
    type=str,
    default="h264_nvenc",
    help="Encoding codec, For x264 class codecs: libx264 (reference but slow), or: h264_nvenc (faster, GPU). For x265 class codecs: libx265 (reference but slow), or: hevc_nvenc(faster, GPU",
    show_default=True,
)
@click.option(
    "--preset",
    "-ps",
    type=str,
    default="slow",
    help="Possible values: ultrafast, superfast, veryfast, faster, fast, medium (default preset), slow, slower, veryslow."
    " A preset is a collection of options that will provide a certain encoding speed to compression ratio."
    " A slower preset will provide better compression (compression is quality per filesize)."
    " This means that, for example, if you target a certain file size or constant bit rate,"
    " you will achieve better quality with a slower preset."
    " Similarly, for constant quality encoding, you will simply save bitrate by choosing a slower preset.",
    show_default=True,
)
@click.option(
    "--quality",
    "-q",
    type=int,
    default=0,
    help="Sets the quality of the video by modulating the bit rate. Set to a higher value for higher quality, or lower (negative ok) value for lower quality. Means different things for different codecs.",
    show_default=True,
)
@click.option(
    "--width",
    "-wi",
    type=int,
    default=0,
    help="Video frame width, recommended values: 1280(HD720), 1920(HD1080), 2048(2K), 3840(UHD-1), 4096(4K), 7680(8K). If zero then original size preserved, "
    "if -1 the best fit is found while allowing some downscaling. Height is automatically determined to preserve aspect ratio,"
    " and is forced to be a multiple of 32.",
    show_default=True,
)
@click.option(
    "--pixelformat",
    "-pf",
    type=str,
    default="yuv420p",
    help="Pixel format, can be yuv420p (default, recommended) or yuv444p (no chrominance downsampling but better compatibility),"
    " or any other allowed ffmpeg & codec pixel format.",
    show_default=True,
)
@click.option(
    "--overwrite", "-w", is_flag=True, help="to force overwrite of target", show_default=True
)  # , help='dataset slice'
def mp4(
    input_path,
    prefix,
    leading,
    extension,
    output_path,
    framerate,
    codec,
    preset,
    quality,
    width,
    pixelformat,
    overwrite,
):
    """Converts a folder of images into an MP4 video file."""

    if output_path is None:
        videofilepath = os.path.basename(os.path.normpath(input_path)) + ".mp4"
    else:
        videofilepath = output_path

    videofilepath = videofilepath.replace("frames_", "")

    if overwrite or not exists(videofilepath):
        with asection(
            f"Converting PNG files at: {input_path}, into MP4 file: {videofilepath}, framerate: {framerate} "
        ):

            scale_option = f"-vf scale={width}:-8:flags=bicubic" if width > 0 else ""
            # black_background_filter = f'-filter_complex "{scale_option};color=black,format={pixelformat}[c];[c][0]scale2ref[c][i];[c][i]overlay=format=auto:shortest=1,setsar=1"'

            # some codec wizardry to be able to effectively modulate quality
            quality = (
                f"-rc vbr -cq {26 - quality} -qmin {26 - quality} -qmax {26 - quality} -b:v 0 "
                if "nvenc" in codec
                else f"-crf {21 - quality}"
            )

            ffmpeg_command = (
                f"ffmpeg -hwaccel auto -framerate {framerate} -start_number 0 -i '{input_path}/{prefix}%0{leading}d.{extension}'  "
                f"-f mp4 -vcodec {codec} -preset {preset}  {quality} -pix_fmt {pixelformat} {scale_option} -y {videofilepath}"
            )
            # -vf  \"crop=576:1240:320:0\"  ""  ,setsar=1:1
            # -pix_fmt {pixelformat} {scale_option}
            # ,setsar=1:1
            # h264_nvenc nvenc_h264 libx264

            aprint(f"Executing command: '{ffmpeg_command}'")

            os.system(ffmpeg_command)
    else:
        aprint(f"Video file: {videofilepath} already exists! use -w option to force overwrite...")


#
# #low quality
# #ffmpeg -r 4 -i /data_fish_TL100_range1200um_step0.31_4um_15ms_dualv_fused_colormax_projection/frame_%05d.png -vcodec mpeg4 -y movie.mp4
#
# #high quality
# ffmpeg -framerate 30 -i data_fish_TL100_range1200um_step0.31_4um_15ms_dualv_fused_stabilized_colormax_projection/frame_%05d.png -f mp4 -vcodec libx264 -preset slow -pix_fmt yuv420p -y movie.mp4
#
# #change resolution
# ffmpeg -i movie.mp4 -s hd1080 movie_1080.mp4
# # or use -vf scale=1920x1080 to set to other scales



from dexp.utils.backends.backend import Backend
from dexp.utils.backends.best_backend import BestBackend
from dexp.utils.backends.cupy_backend import CupyBackend
from dexp.utils.backends.numpy_backend import NumpyBackend



from typing import Optional, Sequence

import numpy
from arbol.arbol import aprint
from skimage.transform import downscale_local_mean

from dexp.datasets import BaseDataset
from dexp.utils.timeit import timeit


def dataset_isonet(
    dataset: BaseDataset,
    path: str,
    channel: Optional[Sequence[str]],
    slicing,
    store: str,
    compression: str,
    compression_level: int,
    overwrite: bool,
    context: str,
    mode: str,
    dxy: float,
    dz: float,
    binning: int,
    sharpening: bool,
    training_tp_index: Optional[int],
    max_epochs: int,
    check: bool,
):
    if channel is None:
        channel = "fused"

    if training_tp_index is None:
        training_tp_index = dataset.nb_timepoints(channel) // 2

    aprint(f"Selected channel {channel}")

    aprint("getting Dask arrays to apply isonet on...")
    array = dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True)

    if slicing is not None:
        aprint(f"Slicing with: {slicing}")
        array = array[slicing]

    aprint(f"Binning image by a factor {binning}...")
    dxy *= binning

    subsampling = dz / dxy
    aprint(f"Parameters: dxy={dxy}, dz={dz}, subsampling={subsampling}")

    psf = numpy.ones((1, 1)) / 1
    print(f"PSF (along xy): {psf}")
    from dexp.processing.isonet import IsoNet

    isonet = IsoNet(context, subsampling=subsampling)

    if "p" in mode:
        training_array_tp = array[training_tp_index].compute()
        training_downscaled_array_tp = downscale_local_mean(training_array_tp, factors=(1, binning, binning))
        print(f"Training image shape: {training_downscaled_array_tp.shape} ")
        isonet.prepare(training_downscaled_array_tp, psf=psf, threshold=0.999)

    if "t" in mode:
        isonet.train(max_epochs=max_epochs)

    if "a" in mode:
        from dexp.datasets import ZDataset

        mode = "w" + ("" if overwrite else "-")
        dest_dataset = ZDataset(path, mode, store, parent=dataset)
        zarr_array = None

        for tp in range(0, array.shape[0] - 1):
            with timeit("Elapsed time: "):

                aprint(f"Processing time point: {tp} ...")
                tp_array = array[tp].compute()

                aprint("Downscaling image...")
                tp_array = downscale_local_mean(tp_array, factors=(1, binning, binning))
                # array_tp_downscaled = zoom(tp_array, zoom=(1, 1.0/binning, 1.0/binning), order=0)

                if sharpening:
                    aprint("Sharpening image...")
                    from dexp.processing.restoration import dehazing

                    tp_array = dehazing(tp_array, mode="hybrid", min=0, max=1024, margin_pad=False)

                aprint("Applying IsoNet to image...")
                tp_array = isonet.apply(tp_array)

                aprint(f"Result: image of shape: {tp_array.shape}, dtype: {tp_array.dtype} ")

                if zarr_array is None:
                    shape = (array.shape[0],) + tp_array.shape
                    aprint(f"Creating Zarr array of shape: {shape} ")
                    zarr_array = dest_dataset.add_channel(
                        name=channel, shape=shape, dtype=array.dtype, codec=compression, clevel=compression_level
                    )

                aprint("Writing image to Zarr file...")
                dest_dataset.write_stack(channel=channel, time_point=tp, stack_array=tp_array)

                aprint(f"Done processing time point: {tp}")

        aprint(dest_dataset.info())
        if check:
            dest_dataset.check_integrity()

        # close destination dataset:
        dest_dataset.close()



from typing import Optional, Sequence

import numpy
from arbol.arbol import aprint, asection
from joblib import Parallel, delayed

from dexp.datasets import BaseDataset
from dexp.utils.misc import compute_num_workers
from dexp.utils.slicing import slice_from_shape


def dataset_copy(
    dataset: BaseDataset,
    dest_path: str,
    channels: Sequence[str],
    slicing,
    store: str = "dir",
    chunks: Optional[Sequence[int]] = None,
    compression: str = "zstd",
    compression_level: int = 3,
    overwrite: bool = False,
    zerolevel: int = 0,
    workers: int = 1,
    workersbackend: Optional[str] = None,
    check: bool = True,
    stop_at_exception: bool = True,
):

    # Create destination dataset:
    from dexp.datasets import ZDataset

    mode = "w" + ("" if overwrite else "-")
    dest_dataset = ZDataset(dest_path, mode, store, parent=dataset)

    # Process each channel:
    for channel in dataset._selected_channels(channels):
        with asection(f"Copying channel {channel}:"):
            array = dataset.get_array(channel)

            aprint(f"Slicing with: {slicing}")
            out_shape, volume_slicing, time_points = slice_from_shape(array.shape, slicing)

            dtype = array.dtype
            dest_dataset.add_channel(
                name=channel, shape=out_shape, dtype=dtype, chunks=chunks, codec=compression, clevel=compression_level
            )

            def process(i):
                tp = time_points[i]
                try:
                    aprint(f"Processing time point: {i} ...")
                    tp_array = array[tp][volume_slicing]
                    if zerolevel != 0:
                        tp_array = numpy.array(tp_array)
                        tp_array = numpy.clip(tp_array, a_min=zerolevel, a_max=None, out=tp_array)
                        tp_array -= zerolevel
                    dest_dataset.write_stack(channel=channel, time_point=i, stack_array=tp_array)
                except Exception as error:
                    aprint(error)
                    aprint(f"Error occurred while copying time point {i} !")
                    import traceback

                    traceback.print_exc()
                    if stop_at_exception:
                        raise error

            if workers == 1:
                for i in range(len(time_points)):
                    process(i)
            else:
                n_jobs = compute_num_workers(workers, len(time_points))

                parallel = Parallel(n_jobs=n_jobs, backend=workersbackend)
                parallel(delayed(process)(i) for i in range(len(time_points)))

    # Dataset info:
    aprint(dest_dataset.info())

    # Check dataset integrity:
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



# flake8: noqa
from arbol import Arbol, asection

from dexp.processing.crop.representative_crop import representative_crop
from dexp.utils.backends import CupyBackend, NumpyBackend


def demo_rep_crop_numpy():
    with NumpyBackend():
        _demo_representative_crop(fast_mode=True)
        _demo_representative_crop(fast_mode=False)


def demo_rep_crop_cupy():
    try:
        with CupyBackend():
            _demo_representative_crop(fast_mode=True)
            _demo_representative_crop(fast_mode=False)
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_representative_crop(fast_mode: bool = True, display: bool = True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Arbol.enable_output = True
    Arbol.set_log_max_depth(5)

    from skimage import data
    from skimage.color import rgb2gray

    image = data.astronaut()
    image = rgb2gray(image)

    with asection(f"Computing crop for image of shape: {image.shape}"):
        crop_size = 16000
        crop = representative_crop(image, crop_size=crop_size, fast_mode=fast_mode, display=False)

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name="image")
            viewer.add_image(crop, name="crop")

    assert crop.size <= int(crop_size * 1.05)


if __name__ == "__main__":
    demo_rep_crop_cupy()
    demo_rep_crop_numpy()



import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import DEFAULT_CLEVEL, DEFAULT_CODEC, DEFAULT_STORE
from dexp.cli.parsing import _get_output_path, _parse_channels, _parse_chunks
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.crop import dataset_crop


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option("--channels", "-c", default=None, help="List of channels, all channels when ommited.")
@click.option(
    "--quantile",
    "-q",
    default=0.99,
    type=float,
    help="Quantile parameter for lower bound of brightness for thresholding.",
    show_default=True,
)
@click.option(
    "--reference-channel",
    "-rc",
    default=None,
    help="Reference channel to estimate cropping. If no provided it picks the first one.",
)
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option("--chunks", "-chk", default=None, help="Dataset chunks dimensions, e.g. (1, 126, 512, 512).")
@click.option(
    "--codec",
    "-z",
    default=DEFAULT_CODEC,
    help="Compression codec: zstd for ’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ",
    show_default=True,
)
@click.option("--clevel", "-l", type=int, default=DEFAULT_CLEVEL, help="Compression level", show_default=True)
@click.option("--overwrite", "-w", is_flag=True, help="Forces overwrite of target", show_default=True)
@click.option(
    "--workers",
    "-wk",
    default=-4,
    help="Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n| ",
    show_default=True,
)  #
@click.option("--check", "-ck", default=True, help="Checking integrity of written file.", show_default=True)  #
def crop(
    input_paths,
    output_path,
    channels,
    quantile,
    reference_channel,
    store,
    chunks,
    codec,
    clevel,
    overwrite,
    workers,
    check,
):

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_crop")
    channels = _parse_channels(input_dataset, channels)
    if reference_channel is None:
        reference_channel = input_dataset.channels()[0]
    chunks = _parse_chunks(chunks)

    with asection(
        f"Cropping from: {input_paths} to {output_path} for channels: {channels}, "
        f"using channel {reference_channel} as a reference."
    ):

        dataset_crop(
            input_dataset,
            output_path,
            channels=channels,
            reference_channel=reference_channel,
            quantile=quantile,
            store=store,
            chunks=chunks,
            compression=codec,
            compression_level=clevel,
            overwrite=overwrite,
            workers=workers,
            check=check,
        )

        input_dataset.close()
        aprint("Done!")



from typing import Optional

import numexpr
import numpy as np

from dexp.utils import xpArray
from dexp.utils.backends import Backend, CupyBackend


def blend_images(array_a: xpArray, array_b: xpArray, blend_map: xpArray, dtype: Optional[np.dtype] = None) -> xpArray:
    """
    Blends two arrays on the basis of a blend map. A value of 1 in the blend map signifies that the value of
    the first array is chosen, a value of 0 means that the value from teh second array is chosen instead.
    For values within ]0,1[ linear blending is performed.

    Parameters
    ----------
    array_a : first array to blend
    array_b : second array to blend
    blend_map : blend map with values within [0, 1]
    dtype   : dtype for returned array

    Returns
    -------
    Blended array.

    """

    if array_a.dtype != array_b.dtype:
        raise ValueError("Two arrays to blend must have same dtype!")

    if array_a.shape != array_b.shape:
        raise ValueError("Two arrays to blend must have same shape!")

    if dtype is None:
        dtype = array_a.dtype

    array_a = Backend.to_backend(array_a, dtype=dtype)
    array_b = Backend.to_backend(array_b, dtype=dtype)
    blend_map = Backend.to_backend(blend_map)

    backend = Backend.current()

    if isinstance(backend, CupyBackend):
        import cupy

        @cupy.fuse()
        def blend_function(a, b, m):
            return a * m + (1 - m) * b

        return blend_function(array_a, array_b, blend_map)

    else:
        a = array_a  # noqa
        b = array_b  # noqa
        m = blend_map  # noqa
        return numexpr.evaluate("a*m+(1-m)*b")



import click
from arbol.arbol import aprint, asection

from dexp.cli.parsing import _parse_channels, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.extract_psf import dataset_extract_psf


@click.command(name="extract-psf")
@click.argument("input_paths", nargs=-1, required=True)
@click.option("--out-prefix-path", "-o", default="psf_", help="Output PSF file prefix")
@click.option(
    "--channels",
    "-c",
    default=None,
    help="list of channels to extract the PSF, each channel will have their separater PSF (i.e., psf_c0.npy, psf_c1.npy",
)
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)
@click.option(
    "--peak-threshold",
    "-pt",
    type=int,
    default=500,
    show_default=True,
    help="Peak valeu threshold for object (PSF) detection. Lower values are less consertaive and will detect more objects.",
)
@click.option(
    "--similarity-threshold",
    "-st",
    type=float,
    default=0.5,
    show_default=True,
    help="Threshold of PSF selection given the similarity (cosine distance) to median PSF.",
)
@click.option("--psf_size", "-ps", type=int, default=35, show_default=True, help="Size (shape) of the PSF")
@click.option(
    "--device",
    "-d",
    type=int,
    default=0,
    help="Sets the CUDA device id, e.g. 0,1,2. It works for only a single device!",
    show_default=True,
)
@click.option("--verbose", "-v", type=bool, is_flag=True, default=False, help="Flag to display intermediated results.")
def extract_psf(
    input_paths,
    out_prefix_path,
    channels,
    slicing,
    peak_threshold,
    similarity_threshold,
    psf_size,
    device,
    verbose,
):
    """Extracts the PSF from beads."""
    input_dataset, input_paths = glob_datasets(input_paths)
    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)

    with asection(
        f"Extracting PSF of dataset: {input_paths}, saving it with prefix: {out_prefix_path}, for channels: {channels}, slicing: {slicing} "
    ):
        aprint(f"Device used: {device}")
        dataset_extract_psf(
            dataset=input_dataset,
            dest_path=out_prefix_path,
            channels=channels,
            slicing=slicing,
            peak_threshold=peak_threshold,
            similarity_threshold=similarity_threshold,
            psf_size=psf_size,
            verbose=verbose,
            device=device,
        )



import os

import click
from arbol.arbol import aprint, asection

from dexp.cli.parsing import _get_output_path
from dexp.utils.robocopy import robocopy


@click.command()
@click.argument("input_path", nargs=1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option(
    "--workers",
    "-wk",
    default=8,
    help="Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n| ",
    show_default=True,
)  #
@click.option(
    "--large_files", "-lf", is_flag=True, help="Set to true to speed up large file transfer", show_default=True
)
def fastcopy(input_path, output_path, workers, large_files):
    """Copies a dataset fast, with no processing, just moves the data as fast as possible. For each operating system it uses the best method."""

    output_path = _get_output_path(input_path, output_path, "_copy")

    if workers == -1:
        workers = max(1, os.cpu_count() // abs(workers))

    with asection(f"Fast copying from: {input_path} to {output_path} "):

        from sys import platform

        if platform == "linux" or platform == "linux2":
            raise NotImplementedError("Fast copy not yet implemented for Linux")
        elif platform == "darwin":
            raise NotImplementedError("Fast copy not yet implemented for OSX")
        elif platform == "win32":
            robocopy(
                input_path, output_path, nb_threads=workers, large_files=large_files or not (".zarr" in input_path)
            )

        aprint("Done!")



from typing import Optional

from dexp.utils import xpArray
from dexp.utils.backends import Backend


def add_noise(
    image: xpArray,
    normalise: bool = False,
    intensity: int = 5,
    variance: float = 0.01,
    seed: Optional[int] = None,
):
    """
    Adds Poisson and Gaussian noise to an image.

    Parameters
    ----------
    image: array
        Image to add noise to.
    intensity: int
        intensity for Possion noise
    variance: float
        Gaussian noise variance.
    clip
    seed

    Returns
    -------

    """

    # Backend:
    xp = Backend.get_xp_module(image)

    # copy image:
    image = image.copy()

    # Normalise:
    if normalise:
        image -= image.min()
        image /= image.max()

    # Sets teh seed:
    if seed is not None:
        seed = abs(seed)
        xp.random.seed(seed)

    # Applies Poisson noise:
    if intensity is not None:
        image = xp.random.poisson(image * intensity) / intensity

    # Applies Gaussian noise:
    image += xp.random.normal(0, variance, size=image.shape)

    # Resets dtype:
    image = image.astype(xp.float32, copy=False)

    # Returns image:
    return image



import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import (
    DEFAULT_CLEVEL,
    DEFAULT_CODEC,
    DEFAULT_STORE,
    DEFAULT_WORKERS_BACKEND,
)
from dexp.cli.parsing import (
    _get_output_path,
    _parse_channels,
    _parse_chunks,
    _parse_slicing,
)
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.copy import dataset_copy


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option("--channels", "-c", default=None, help="List of channels, all channels when ommited.")
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="Dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option("--chunks", "-chk", default=None, help="Dataset chunks dimensions, e.g. (1, 126, 512, 512).")
@click.option(
    "--codec",
    "-z",
    default=DEFAULT_CODEC,
    help="Compression codec: zstd for ’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ",
    show_default=True,
)
@click.option("--clevel", "-l", type=int, default=DEFAULT_CLEVEL, help="Compression level", show_default=True)
@click.option("--overwrite", "-w", is_flag=True, help="Forces overwrite of target", show_default=True)
@click.option(
    "--zerolevel",
    "-zl",
    type=int,
    default=0,
    help="‘zero-level’ i.e. the pixel values in the restoration (to be substracted)",
    show_default=True,
)  #
@click.option(
    "--workers",
    "-wk",
    default=-4,
    help="Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n| ",
    show_default=True,
)  #
@click.option(
    "--workersbackend",
    "-wkb",
    type=str,
    default=DEFAULT_WORKERS_BACKEND,
    help="What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ",
    show_default=True,
)  #
@click.option("--check", "-ck", default=True, help="Checking integrity of written file.", show_default=True)  #
def copy(
    input_paths,
    output_path,
    channels,
    slicing,
    store,
    chunks,
    codec,
    clevel,
    overwrite,
    zerolevel,
    workers,
    workersbackend,
    check,
):
    """Copies a dataset, channels can be selected, cropping can be performed, compression can be changed, ..."""

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_copy")
    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)
    chunks = _parse_chunks(chunks)

    with asection(f"Copying from: {input_paths} to {output_path} for channels: {channels}, slicing: {slicing} "):
        dataset_copy(
            input_dataset,
            output_path,
            channels=channels,
            slicing=slicing,
            store=store,
            chunks=chunks,
            compression=codec,
            compression_level=clevel,
            overwrite=overwrite,
            zerolevel=zerolevel,
            workers=workers,
            workersbackend=workersbackend,
            check=check,
        )

        input_dataset.close()
        aprint("Done!")



from dexp.datasets.base_dataset import BaseDataset
from dexp.datasets.clearcontrol_dataset import CCDataset
from dexp.datasets.joined_dataset import JoinedDataset
from dexp.datasets.tiff_dataset import TIFDataset
from dexp.datasets.zarr_dataset import ZDataset



import math
import os
from os.path import exists, join

import click
import numpy
from arbol.arbol import aprint, asection, section

from dexp.cli.parsing import _parse_channels, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets


@click.command()
@click.argument("input_paths", nargs=-1)
@click.option(
    "--output_path", "-o", default=None, help="Output folder to store rendered PNGs. Default is: frames_<channel_name>"
)
@click.option("--channels", "-c", default=None, help="list of channels to color, all channels when ommited.")
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z).",
)
@click.option(
    "--overwrite", "-w", is_flag=True, help="to force overwrite of target", show_default=True
)  # , help='dataset slice'
@click.option("--aspect", "-a", type=float, default=4, help="sets aspect ratio e.g. 4", show_default=True)
@click.option(
    "--colormap",
    "-cm",
    type=str,
    default="magma",
    help="sets colormap, e.g. viridis, gray, magma, plasma, inferno ",
    show_default=True,
)
@click.option(
    "--rendersize",
    "-rs",
    type=int,
    default=2048,
    help="Sets the frame color size. i.e. -ws 400 sets the window to 400x400",
    show_default=True,
)
@click.option(
    "--clim",
    "-cl",
    type=str,
    default="0,1024",
    help="Sets the contrast limits, i.e. -cl 0,1000 sets the contrast limits to [0,1000]",
)
@click.option(
    "--options",
    "-opt",
    type=str,
    default=None,
    help="Render options, e.g: 'gamma=1.2,box=True'. Important: no white spaces!!! Complete list with defaults will be displayed on first run",
)
def volrender(
    input_paths,
    output_path,
    channels=None,
    slicing=None,
    overwrite=False,
    aspect=None,
    colormap="viridis",
    rendersize=1536,
    clim=None,
    options=None,
):
    """Renders 3D volumetric video"""

    input_dataset, input_paths = glob_datasets(input_paths)
    channels = _parse_channels(input_dataset, channels)
    slicing = _parse_slicing(slicing)

    aprint(f"Volumetric rendering of: {input_paths} to {output_path} for channels: {channels}, slicing: {slicing} ")

    for channel in channels:

        with asection(f"Channel '{channel}' shape: {input_dataset.shape(channel)}:"):
            aprint(input_dataset.info(channel))

        array = input_dataset.get_array(channel, wrap_with_dask=True)
        dtype = array.dtype

        if slicing:
            array = array[slicing]

        aprint(f"Rendering array of shape={array.shape} and dtype={array.dtype} for channel '{channel}'.")

        if clim is None:
            aprint("Computing min and max from first stack...")
            first_stack = numpy.array(input_dataset.get_stack(channel, 0, per_z_slice=False))
            min_value = max(0, first_stack.min() - 100)
            max_value = first_stack.max() + 100
            aprint(f"min={min_value} and max={max_value}.")
        else:
            aprint(f"provided min and max for contrast limits: {clim}")
            min_value, max_value = (float(strvalue) for strvalue in clim.split(","))

        aprint(f"Provided rendering options: {options}")
        options = dict(item.split("=") for item in options.split(",")) if options is not None else dict()

        def str2bool(v):
            return v.lower() in ("yes", "true", "t", "1")

        nbtp = array.shape[0]
        nbframes = int(options["nbframes"]) if "nbframes" in options else 1010

        skip = int(options["skip"]) if "skip" in options else 1
        cut = str(options["cut"]) if "cut" in options else "nocut"
        cutpos = float(options["cutpos"]) if "cutpos" in options else 0
        cutspeed = float(options["cutspeed"]) if "cutspeed" in options else 0
        irot = options["irot"] if "irot" in options else ""
        raxis = options["raxis"] if "raxis" in options else "y"
        rstart = float(options["rstart"]) if "rstart" in options else 0
        rspeed = float(options["rspeed"]) if "rspeed" in options else 0.15
        tstart = int(options["tstart"]) if "tstart" in options else 0
        tspeed = float(options["tspeed"]) if "tspeed" in options else 0.5
        gamma = float(options["gamma"]) if "gamma" in options else 0.9
        zoom = float(options["zoom"]) if "zoom" in options else 1.45
        alpha = float(options["alpha"]) if "alpha" in options else 0.3
        box = str2bool(options["box"]) if "box" in options else False
        maxsteps = int(options["maxsteps"]) if "maxsteps" in options else 1024
        norm = str2bool(options["norm"]) if "norm" in options else True
        normrange = float(options["normrange"]) if "normrange" in options else 1024
        videofilename = options["name"] if "name" in options else "video.mp4"

        with asection("Options:"):
            aprint(f"Video filename          : {videofilename}")
            aprint(f"Number of time points   : {nbtp}")
            aprint(f"Number of frames        : {nbframes}  \toption: nbframes: \tint")
            aprint(f"Render one frame every  : {skip}      \toption: skip:     \tint")
            aprint(
                f"Volume cutting          : {cut}       \toption: cut:      \t[left, right, top, bottom, front, back, none]"
            )
            aprint(f"Volume cutting position : {cutpos}    \toption: cutpos:   \tfloat")
            aprint(f"Volume cutting speed    : {cutspeed}  \toption: cutspeed: \tfloat")
            aprint(f"Initial time point      : {tstart}    \toption: tstart:   \tint")
            aprint(f"Time    speed           : {tspeed}    \toption: tspeed:   \tfloat")
            aprint(
                f"Initial rotation        : {irot}      \toption: irot:     \t xxyzzz -> 90deg along x, 45deg along y, 135deg along z "
            )
            aprint(f"Rotation axis           : {raxis}     \toption: raxis:    \t[x,y,z]")
            aprint(f"Initial rotation angle  : {rstart}    \toption: rstart:   \tfloat")
            aprint(f"Rotation speed          : {rspeed}    \toption: rspeed:   \tfloat")
            aprint(f"Gamma                   : {gamma}     \toption: gamma:    \tfloat")
            aprint(f"Zoom                    : {zoom}      \toption: zoom:     \tfloat")
            aprint(f"Alpha                   : {alpha}     \toption: alpha:    \tfloat")
            aprint(f"box?                    : {box}       \toption: box:      \tbool (true/false)")
            aprint(f"normalisation           : {norm}      \toption: norm:     \tbool (true/false)")
            aprint(f"normalisation range     : {normrange} \toption: normrange:\tfloat")
            aprint(f"Max steps for vol color: {maxsteps}  \toption: maxsteps: \tint")

        if output_path is None:
            output_path = f"frames_{channel}"
        os.makedirs(output_path, exist_ok=True)
        from spimagine import DataModel, NumpyData, volshow

        aprint("Opening spimagine...")
        import spimagine

        spimagine.config.__DEFAULTMAXSTEPS__ = maxsteps
        spimagine.config.__DEFAULT_TEXTURE_WIDTH__ = rendersize

        datamodel = DataModel(NumpyData(array[0].compute()))
        aprint("Creating Spimagine window... (you can minimise but don't close!)")
        win = volshow(datamodel, stackUnits=(1.0, 1.0, aspect), autoscale=False, show_window=True)
        win.resize(rendersize + 32, rendersize + 32)
        win.showMinimized()

        while section("Rendering:"):
            for i in range(0, nbframes, skip):
                aprint("______________________________________________________________________________")
                aprint(f"Frame     : {i}")

                tp = tstart + int(tspeed * i)
                if tp >= nbtp:
                    break
                aprint(f"Time point: {tp}")

                angle = rstart + rspeed * i
                aprint(f"Angle     : {angle}")

                effcutpos = cutpos + cutspeed

                filename = join(output_path, f"frame_{i:05}.png")

                if overwrite or not exists(filename):

                    aprint("Loading stack...")
                    stack = array[int(tp)].compute()

                    if norm:
                        aprint("Computing percentile...")
                        rmax = numpy.percentile(stack[::8].astype(numpy.float32), q=99.99).astype(numpy.float32)
                        # rmax = numpy.max(stack[::8]).astype(numpy.float32)
                        aprint(f"rmax={rmax}")

                        aprint("Normalising...")
                        norm_max_value = normrange
                        norm_min_value = 64.0
                        # stack = norm_min_value+stack*((norm_max_value-norm_min_value)/rmax)
                        stack = stack * numpy.array((norm_max_value - norm_min_value) / rmax, dtype=numpy.float32)
                        stack += numpy.array(norm_min_value, dtype=dtype)
                        stack = stack.astype(dtype)

                    # print("Opening spimagine...")
                    # win = volshow(stack, stackUnits=(1., 1., aspect), autoscale=False, show_window=True)

                    aprint("Loading stack into Spimagine...")
                    datamodel.setContainer(NumpyData(stack))
                    win.setModel(datamodel)

                    aprint("Setting rendering parameters...")
                    if colormap in spimagine.config.__COLORMAPDICT__:
                        win.set_colormap(colormap)
                    else:
                        from matplotlib import colors

                        rgb = colors.to_rgba(colormap)[:3]
                        aprint(f"Turning the provided color: {colormap} = {rgb} into a colormap.")
                        win.set_colormap_rgb(rgb)

                    win.transform.setStackUnits(1.0, 1.0, aspect)
                    win.transform.setGamma(gamma)
                    win.transform.setMin(min_value)
                    win.transform.setMax(max_value)
                    win.transform.setZoom(zoom)
                    win.transform.setAlphaPow(alpha)
                    win.transform.setBox(box)

                    if cut == "left":
                        win.transform.setBounds(effcutpos, 1, -1, 1, -1, 1)
                    elif cut == "right":
                        win.transform.setBounds(-1, effcutpos, -1, 1, -1, 1)
                    elif cut == "top":
                        win.transform.setBounds(-1, 1, effcutpos, 1, -1, 1)
                    elif cut == "bottom":
                        win.transform.setBounds(-1, 1, -1, effcutpos, -1, 1)
                    elif cut == "front":
                        win.transform.setBounds(-1, 1, -1, 1, effcutpos, 1)
                    elif cut == "back":
                        win.transform.setBounds(-1, 1, -1, 1, -1, effcutpos)
                    elif cut == "centerx":
                        win.transform.setBounds(-0.25 - effcutpos, 0.25 + effcutpos, -1, 1, -1, 1)
                    elif cut == "centery":
                        win.transform.setBounds(-1, 1, -0.25 - effcutpos, 0.25 + effcutpos, -1, 1)
                    elif cut == "centerz":
                        win.transform.setBounds(-1, 1, -1, 1, -0.25 - effcutpos, 0.25 + effcutpos)
                    elif cut == "none":
                        win.transform.setBounds(-1, 1, -1, 1, -1, 1)

                    win.transform.setRotation(0, 1, 0, 0)

                    for character in irot:
                        if character == "x":
                            aprint(f"Rotating along x axis by 45 deg (prev quatRot={win.transform.quatRot})")
                            win.transform.addRotation(0.5 * math.pi / 4, 1, 0, 0)
                        elif character == "y":
                            aprint(f"Rotating along y axis by 45 deg (prev quatRot={win.transform.quatRot})")
                            win.transform.addRotation(0.5 * math.pi / 4, 0, 1, 0)
                        elif character == "z":
                            aprint(f"Rotating along z axis by 45 deg (prev quatRot={win.transform.quatRot})")
                            win.transform.addRotation(0.5 * math.pi / 4, 0, 0, 1)
                    aprint(f"Rotation after initial axis rotation: {win.transform.quatRot}")

                    if "x" in raxis:
                        aprint(f"Rotating along x axis by {angle} deg (prev quatRot={win.transform.quatRot})")
                        win.transform.addRotation(0.5 * angle * (math.pi / 180), 1, 0, 0)
                    if "y" in raxis:
                        aprint(f"Rotating along y axis by {angle} deg (prev quatRot={win.transform.quatRot})")
                        win.transform.addRotation(0.5 * angle * (math.pi / 180), 0, 1, 0)
                    if "z" in raxis:
                        aprint(f"Rotating along z axis by {angle} deg (prev quatRot={win.transform.quatRot})")
                        win.transform.addRotation(0.5 * angle * (math.pi / 180), 0, 0, 1)

                    aprint(f"Final rotation: {win.transform.quatRot}")

                    aprint(f"Saving frame: {filename}")
                    win.saveFrame(filename)

    win.closeMe()
    input_dataset.close()
    aprint("Done!")

    raise SystemExit
    import sys

    sys.exit()



import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import DEFAULT_STORE
from dexp.cli.parsing import _get_output_path, _parse_channels
from dexp.datasets.open_dataset import glob_datasets


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option("--channels", "-c", default=None, help="List of channels, all channels when ommited.")
@click.option(
    "--rename",
    "-rc",
    default=None,
    help="You can rename channels: e.g. if channels are ‘channel1,anotherc’ then ‘gfp,rfp’ would rename the ‘channel1’ channel to ‘gfp’, and ‘anotherc’ to ‘rfp’ ",
)
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option("--overwrite", "-w", is_flag=True, help="Forces overwrite of target", show_default=True)
@click.option(
    "--projection", "-p/-np", is_flag=True, default=True, help="If flags should be copied.", show_default=True
)
def add(input_paths, output_path, channels, rename, store, overwrite, projection):
    """Adds the channels selected from INPUT_PATHS to the given output dataset (created if not existing)."""

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_add")
    channels = _parse_channels(input_dataset, channels)

    if rename is None:
        rename = input_dataset.channels()
    else:
        rename = rename.split(",")

    with asection(f"Adding channels: {channels} from: {input_paths} to {output_path}, with new names: {rename}"):
        input_dataset.add_channels_to(
            output_path,
            channels=channels,
            rename=rename,
            store=store,
            overwrite=overwrite,
            add_projections=projection,
        )

        input_dataset.close()
        aprint("Done!")



from functools import partial
from typing import Iterable, Optional, Sequence, Tuple, Union

import numpy as np
from scipy.signal._signaltools import _centered

from dexp.processing.crop.representative_crop import representative_crop
from dexp.processing.denoising.j_invariance import calibrate_denoiser
from dexp.utils import dict_or, xpArray
from dexp.utils.backends import Backend, CupyBackend


def calibrate_denoise_butterworth(
    image: xpArray,
    mode: str = "full",
    axes: Optional[Tuple[int, ...]] = None,
    padding: int = 32,
    min_freq: float = 0.001,
    max_freq: float = 1.0,
    num_freq: int = 32,
    min_order: float = 0.5,
    max_order: float = 6.0,
    num_order: int = 32,
    crop_size_in_voxels: Optional[int] = 256 ** 3,
    display: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the Butterworth denoiser for the given image and returns the optimal
    parameters obtained using the N2S loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate Sobolev denoiser for.

    mode: str
        Possible modes are: 'isotropic' for isotropic meaning only one
        frequency cut-off is calibrated for all axes , 'xy-z' for 3D stacks where
        the cut-off frequency for the x and y axes is the same but different for
        the z axis, and 'full' for which all frequency cut-offs are different.

    axes: Optional[Tuple[int,...]]
        Axes over which to apply low-pass filtering.
        (advanced)

    padding: int
        Amount of padding to be added to avoid edge effects.
        (advanced)

    min_freq: float
        Minimum cutoff frequency to use for calibration. Must be within [0,1],
        typically close to zero.
        (advanced)

    max_freq: float
        Maximum cutoff frequency to use for calibration. Must be within [0,1],
        typically close to one.
        (advanced)

    num_freq: int
        Number of frequencies to use for calibration.

    min_order: float
        Minimal order for the Butterworth filter to use for calibration.
        (advanced)

    max_order: float
        Maximal order for the Butterworth filter to use for calibration.
        (advanced)

    num_order: int
        Number of orders to use for calibration.

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate denoiser.
        (advanced)

    display_images: bool
        When True the denoised images encountered during optimisation are shown.
        (advanced)

    other_fixed_parameters: dict
        Any other fixed parameters. (advanced)

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """

    # Backend:
    xp = Backend.get_xp_module(image)

    # Convert image to float if needed:
    image = image.astype(dtype=xp.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels, display=False)

    # ranges:
    freq_cutoff_range = (
        min_freq,
        max_freq,
        (max_freq - min_freq) / num_freq,
    )  # numpy.arange(min_freq, max_freq, (max_freq-min_freq)/num_freq)
    order_range = (
        min_order,
        max_order,
        (max_order - min_order) / num_order,
    )  # numpy.arange(min_order, max_order, (max_order-min_order)/num_order)

    # Combine fixed parameters:
    other_fixed_parameters = dict_or(
        other_fixed_parameters,
        dict(axes=axes),
    )

    if mode == "isotropic":
        # Partial function:
        _denoise_butterworth = partial(_apply_butterworth, **other_fixed_parameters, out_shape=crop.shape)

        # Parameters to test when calibrating the denoising algorithm
        parameter_ranges = {"freq_cutoff": freq_cutoff_range, "order": order_range}

    elif mode == "xy-z" and image.ndim == 3:
        # Partial function with parameter impedance match:
        def _denoise_butterworth(*args, **kwargs):
            freq_cutoff_xy = kwargs.pop("freq_cutoff_xy")
            freq_cutoff_z = kwargs.pop("freq_cutoff_z")
            _freq_cutoff = (freq_cutoff_xy, freq_cutoff_xy, freq_cutoff_z)
            return _apply_butterworth(
                *args,
                freq_cutoff=_freq_cutoff,
                out_shape=crop.shape,
                **dict_or(kwargs, other_fixed_parameters),
            )

        # Parameters to test when calibrating the denoising algorithm
        parameter_ranges = {
            "freq_cutoff_xy": freq_cutoff_range,
            "freq_cutoff_z": freq_cutoff_range,
            "order": order_range,
        }

    elif mode == "full":
        # Partial function with parameter impedance match:
        def _denoise_butterworth(*args, **kwargs):
            _freq_cutoff = tuple(kwargs.pop(f"freq_cutoff_{i}") for i in range(image.ndim))
            return _apply_butterworth(
                *args,
                out_shape=crop.shape,
                freq_cutoff=_freq_cutoff,
                **dict_or(kwargs, other_fixed_parameters),
            )

        # Parameters to test when calibrating the denoising algorithm
        parameter_ranges = {f"freq_cutoff_{i}": freq_cutoff_range for i in range(image.ndim)}
        parameter_ranges["order"] = order_range

    else:
        raise ValueError(f"Unsupported denoising mode: {mode}")

    # Calibrate denoiser
    best_parameters = dict_or(
        calibrate_denoiser(
            crop,
            _denoise_butterworth,
            denoise_parameters=parameter_ranges,
            setup_function=partial(_setup_butterworth_denoiser, axes=axes, padding=padding),
            mode="shgo+lbfgs",
            max_evaluations=1000,
            display=display,
        ),
        other_fixed_parameters,
    )

    if mode == "full":
        # We need to adjust a bit the type of parameters passed to the denoising function:
        freq_cutoff = tuple(best_parameters.pop(f"freq_cutoff_{i}") for i in range(image.ndim))
        best_parameters = dict_or(best_parameters, {"freq_cutoff": freq_cutoff})

    elif mode == "xy-z":
        # We need to adjust a bit the type of parameters passed to the denoising function:
        freq_cutoff_xy = best_parameters.pop("freq_cutoff_xy")
        freq_cutoff_z = best_parameters.pop("freq_cutoff_z")
        freq_cutoff = (freq_cutoff_z, freq_cutoff_xy, freq_cutoff_xy)
        best_parameters = dict_or(best_parameters, {"freq_cutoff": freq_cutoff})

    if isinstance(Backend.current(), CupyBackend):
        # Free plane cache to avoid running out of memory
        from cupy.fft.config import get_plan_cache

        get_plan_cache().clear()

    return denoise_butterworth, best_parameters


def denoise_butterworth(
    image,
    axes: Optional[Tuple[int, ...]] = None,
    freq_cutoff: Union[float, Sequence[float]] = 0.5,
    order: float = 1,
    padding: int = 32,
):
    """
    Denoises the given image by applying a configurable <a
    href="https://en.wikipedia.org/wiki/Butterworth_filter">Butterworth
    lowpass filter</a>. Remarkably good when your signal
    does not have high-frequencies beyond a certain cutoff frequency.
    This is probably the first algorithm that should be tried of all
    currently available in Aydin. It is actually quite impressive how
    well this performs in practice. If the signal in your images is
    band-limited as is often the case for microscopy images, this
    denoiser will work great.
    \n\n
    Note: We recommend applying a variance stabilisation transform
    to improve results for images with non-Gaussian noise.

    Parameters
    ----------
    image: ArrayLike
        Image to be denoised

    axes: Optional[Tuple[int,...]]
        Axes over which to apply lowpass filtering.

    freq_cutoff: Union[float, Sequence[float]]
        Single or sequence cutoff frequency, must be within [0, 1]

    order: float
        Filter order, typically an integer above 1.

    padding: int
        Amount of padding to be added to avoid edge effects.

    Returns
    -------
    Denoised image

    """
    if not isinstance(freq_cutoff, Iterable):
        freq_cutoff = tuple((freq_cutoff,) * image.ndim)

    # Computes fft and grid distance map of input image
    data = _setup_butterworth_denoiser(image, axes=axes, padding=padding)
    return _apply_butterworth(data, axes, freq_cutoff, order, image.shape)


def _apply_butterworth(
    data: Tuple[xpArray, Sequence[xpArray]],
    axes: Optional[Sequence[int]],
    freq_cutoff: Tuple[float],
    order: float,
    out_shape: Tuple[int],
) -> xpArray:
    """
    Applies the butterworth filter to a pre computed image in the freq. domain
    and a grid for each axes for distance map computation.

    Parameters
    ----------
    data : Tuple[xpArray, Sequence[xpArray]]
        Image in the freq. domain and a grid for each axes.
    axes : Optional[Sequence[int]]
        Axes selected for filtering
    freq_cutoff : Tuple[float]
        Freq. cutoff parameter.
    order : float
        Butterworth order parameter.
    out_shape : Tuple[int]
        Original image input shape, used to crop the data after the inverse FFT.

    Returns
    -------
    xpArray
        Butterworth filtered image.
    """

    image_f, grid = data

    dist = np.zeros_like(image_f, dtype=np.float32)
    for axis, fc in zip(grid, freq_cutoff):
        dist += np.square(axis / fc)

    # Apply filter:
    image_f = _butterworth_filter(image_f, dist, order)

    # Shift back:
    image_f = np.fft.ifftshift(image_f, axes=axes)

    # Back in real space:
    denoised = np.real(np.fft.ifftn(image_f, axes=axes))

    # Crop to remove padding:
    denoised = _centered(denoised, out_shape)

    return denoised


def _setup_butterworth_denoiser(
    image: xpArray, axes: Optional[Tuple[int, ...]], padding: int
) -> Tuple[xpArray, xpArray]:
    """Pre computes the butterworth forward step.

    Parameters
    ----------
    image : xpArray
        Input image.

    axes: Optional[Tuple[int,...]]
        Axes over which to apply lowpass filtering.

    padding: int
        Padding to be added to avoid edge effects.

    Returns
    -------
    Tuple[xpArray, xpArray]
        Input image in the freq. domain and the distance map.
    """
    xp = Backend.get_xp_module(image)

    # Convert image to float if needed:
    image = image.astype(dtype=np.float32)

    # Default axes:
    if axes is None:
        axes = tuple(range(image.ndim))

    # Selected axes:
    selected_axes = tuple((a in axes) for a in range(image.ndim))

    # First we need to pad the image.
    # By how much? this depends on how much low filtering we need to do:
    pad_width = tuple((padding, padding) if selected else 0 for selected in selected_axes)

    # pad image:
    image = np.pad(image, pad_width=pad_width, mode="reflect")

    # Move to frequency space:
    image_f = np.fft.fftn(image, axes=axes)

    # Center frequencies:
    image_f = np.fft.fftshift(image_f, axes=axes)

    # Computes grid for squared distance
    axis_grid = tuple((xp.linspace(-1, 1, s) if sa else xp.zeros((s,))) for sa, s in zip(selected_axes, image.shape))
    f = xp.meshgrid(*axis_grid, indexing="ij")

    return image_f, f


def _butterworth_filter(image_f: xpArray, f: xpArray, order: float) -> xpArray:
    return image_f / np.sqrt(1 + f ** order)



import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import (
    DEFAULT_CLEVEL,
    DEFAULT_CODEC,
    DEFAULT_STORE,
    DEFAULT_WORKERS_BACKEND,
)
from dexp.cli.parsing import _get_output_path, _parse_channels, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.stabilize import dataset_stabilize


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option("--channels", "-c", default=None, help="List of channels, all channels when omitted.")
@click.option(
    "--reference-channel", "-rc", default=None, help="Reference channel for single stabilization model computation."
)
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="Dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option(
    "--codec",
    "-z",
    default=DEFAULT_CODEC,
    help="Compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ",
    show_default=True,
)
@click.option("--clevel", "-l", type=int, default=DEFAULT_CLEVEL, help="Compression level", show_default=True)
@click.option("--overwrite", "-w", is_flag=True, help="Forces overwrite of target", show_default=True)
@click.option(
    "--maxrange",
    "-mr",
    type=int,
    default=7,
    help="Maximal distance, in time points, between pairs of images to registrate.",
    show_default=True,
)
@click.option(
    "--minconfidence",
    "-mc",
    type=float,
    default=0.5,
    help="Minimal confidence for registration parameters, if below that level the registration parameters for previous time points is used.",
    show_default=True,
)  #
@click.option(
    "--com/--no-com",
    type=bool,
    default=False,
    help="Enable center of mass fallback when standard registration fails.",
    show_default=True,
)
@click.option(
    "--quantile",
    "-q",
    type=float,
    default=0.5,
    help="Quantile to cut-off background in center-of-mass calculation.",
    show_default=True,
)
@click.option("--tolerance", "-t", type=float, default=1e-7, help="Tolerance for linear solver.", show_default=True)
@click.option(
    "--ordererror", "-oe", type=float, default=2.0, help="Order for linear solver error term.", show_default=True
)
@click.option(
    "--orderreg", "-or", type=float, default=1.0, help="Order for linear solver regularisation term.", show_default=True
)
@click.option(
    "--alphareg",
    "-or",
    type=float,
    default=1e-4,
    help="Multiplicative coefficient for regularisation term.",
    show_default=True,
)
@click.option(
    "--pcsigma",
    "-rs",
    type=float,
    default=2,
    help="Sigma for Gaussian smoothing of phase correlogram, zero to disable.",
    show_default=True,
)
@click.option(
    "--dsigma",
    "-ds",
    type=float,
    default=1.5,
    help="Sigma for Gaussian smoothing (crude denoising) of input images, zero to disable.",
    show_default=True,
)
@click.option(
    "--logcomp",
    "-lc",
    type=bool,
    default=True,
    help="Applies the function log1p to the images to compress high-intensities (usefull when very (too) bright structures are present in the images, such as beads.",
    show_default=True,
)
@click.option(
    "--edgefilter",
    "-ef",
    type=bool,
    default=False,
    help="Applies sobel edge filter to input images.",
    show_default=True,
)
@click.option(
    "--detrend",
    "-dt",
    type=bool,
    is_flag=True,
    default=False,
    help="Remove linear trend from stabilization result",
    show_default=True,
)
@click.option(
    "--maxproj/--no-maxproj",
    "-mp/-nmp",
    type=bool,
    default=True,
    help="Registers using only the maximum intensity projection from each stack.",
    show_default=True,
)
@click.option(
    "--model-input-path",
    "-mi",
    type=str,
    default=None,
    help="Path to pre-computed model for image registration",
)
@click.option(
    "--model-output-path",
    "-mo",
    type=str,
    default="stabilization_model.json",
    show_default=True,
    help="Output path for computed registration model",
)
@click.option(
    "--workers",
    "-k",
    type=int,
    default=-4,
    help="Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n|. Be careful, starting two many workers is know to cause trouble (unfortunately unclear why!).",
    show_default=True,
)
@click.option(
    "--workersbackend",
    "-wkb",
    type=str,
    default=DEFAULT_WORKERS_BACKEND,
    help="What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ",
    show_default=True,
)  #
@click.option("--device", "-d", type=int, default=0, help="Sets the CUDA devices id, e.g. 0,1,2", show_default=True)  #
@click.option("--check", "-ck", default=True, help="Checking integrity of written file.", show_default=True)  #
def stabilize(
    input_paths,
    output_path,
    channels,
    reference_channel,
    slicing,
    store,
    codec,
    clevel,
    overwrite,
    maxrange,
    minconfidence,
    com,
    quantile,
    tolerance,
    ordererror,
    orderreg,
    alphareg,
    pcsigma,
    dsigma,
    logcomp,
    edgefilter,
    detrend,
    maxproj,
    model_input_path,
    model_output_path,
    workers,
    workersbackend,
    device,
    check,
):
    """Stabilises dataset against translations across time."""

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_stabilized")

    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)

    with asection(
        f"Stabilizing dataset(s): {input_paths}, saving it at: {output_path}, for channels: {channels}, slicing: {slicing} "
    ):
        dataset_stabilize(
            input_dataset,
            output_path,
            channels=channels,
            model_output_path=model_output_path,
            model_input_path=model_input_path,
            reference_channel=reference_channel,
            slicing=slicing,
            zarr_store=store,
            compression_codec=codec,
            compression_level=clevel,
            overwrite=overwrite,
            max_range=maxrange,
            min_confidence=minconfidence,
            enable_com=com,
            quantile=quantile,
            tolerance=tolerance,
            order_error=ordererror,
            order_reg=orderreg,
            alpha_reg=alphareg,
            phase_correlogram_sigma=pcsigma,
            denoise_input_sigma=dsigma,
            log_compression=logcomp,
            edge_filter=edgefilter,
            detrend=detrend,
            maxproj=maxproj,
            workers=workers,
            workers_backend=workersbackend,
            device=device,
            check=check,
            debug_output="stabilization",
        )

        input_dataset.close()
        aprint("Done!")



import ctypes
import os
from typing import Optional


def set_mkl_threads(num_threads: Optional[int] = None) -> None:
    try:
        if num_threads is None:
            import multiprocessing

            num_threads = multiprocessing.cpu_count() // 2 - 1

        try:
            import mkl

            mkl.set_num_threads(num_threads)
            return 0

        except Exception:
            pass
            # import traceback
            # traceback.print_exc()
            # print(e)

        for name in ["libmkl_rt.so", "libmkl_rt.dylib", "mkl_Rt.dll"]:
            try:
                mkl_rt = ctypes.CDLL(name)
                mkl_rt.mkl_set_num_threads(ctypes.byref(ctypes.c_int(num_threads)))
                return 0

            except Exception:
                pass
                # import traceback
                # traceback.print_exc()
                # print(e)

        os.environ["OMP_NUM_THREADS"] = f"{num_threads}"  # export OMP_NUM_THREADS=4
        os.environ["OPENBLAS_NUM_THREADS"] = f"{num_threads}"  # export OPENBLAS_NUM_THREADS=4
        os.environ["MKL_NUM_THREADS"] = f"{num_threads}"  # export MKL_NUM_THREADS=6
        os.environ["VECLIB_MAXIMUM_THREADS"] = f"{num_threads}"  # export VECLIB_MAXIMUM_THREADS=4
        os.environ["NUMEXPR_NUM_THREADS"] = f"{num_threads}"  # export NUMEXPR_NUM_THREADS=6

    except Exception:
        pass
        # import traceback
        # traceback.print_exc()
        # print(e)


# set_mkl_threads(8)



from dexp.processing.denoising.demo.demo_2D_butterworth import _demo_butterworth
from dexp.utils.testing.testing import execute_both_backends


@execute_both_backends
def test_butterworth():
    assert _demo_butterworth(display=False) >= 0.608 - 0.03



from typing import Optional, Sequence, Tuple

import numpy as np
import zarr
from arbol.arbol import aprint, asection
from joblib import Parallel, delayed
from numpy.typing import ArrayLike
from scipy import ndimage as ndi

from dexp.datasets import BaseDataset
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.utils.backends import Backend, BestBackend
from dexp.utils.misc import compute_num_workers


def _estimate_crop(array: ArrayLike, quantile: float = 0.99) -> Sequence[Tuple[int]]:
    window_size = 31
    step = 4
    shape = array.shape
    with BestBackend():
        xp = Backend.get_xp_module()
        array = Backend.to_backend(array[::step, ::step, ::step], dtype=xp.float16)
        array = xp.clip(array - xp.mean(array), 0, None)  # removing background noise
        kernel = xp.ones((window_size, window_size, window_size)) / (window_size ** 3)
        kernel = kernel.astype(xp.float16)
        array = fft_convolve(array, kernel, in_place=True)
        lower = xp.quantile(array, quantile)
        aprint("Estimated lower threshold", lower)
        array = array > lower
        array, _ = ndi.label(Backend.to_numpy(array))
        slices = ndi.find_objects(array)

    largest_slice = None
    largest_size = 0
    for slicing in slices:
        size = np.prod(tuple((s.stop - s.start) for s in slicing))
        if size > largest_size:
            largest_slice = slicing
            largest_size = size

    if largest_slice is None:
        raise RuntimeError("Could not detect any objects")

    return tuple(
        (s.start * step, min(s.stop * step, d))  # fixing possible mismatch due to step
        for s, d in zip(largest_slice, shape)
    )


def compute_crop_slicing(array: zarr.Array, time_points: Sequence[int], quantile: float) -> Sequence[slice]:
    # N x D x 2
    ranges = np.array(tuple(_estimate_crop(array[t], quantile) for t in time_points))
    lower = np.min(ranges[..., 0], axis=0)
    upper = np.max(ranges[..., 1], axis=0)
    return tuple(slice(int(l), int(u)) for l, u in zip(lower, upper))


def dataset_crop(
    dataset: BaseDataset,
    dest_path: str,
    reference_channel: str,
    channels: Sequence[str],
    quantile: float,
    store: str = "dir",
    chunks: Optional[Sequence[int]] = None,
    compression: str = "zstd",
    compression_level: int = 3,
    overwrite: bool = False,
    workers: int = 1,
    check: bool = True,
    stop_at_exception: bool = True,
):

    # Create destination dataset:
    from dexp.datasets import ZDataset

    mode = "w" + ("" if overwrite else "-")
    dest_dataset = ZDataset(dest_path, mode, store, parent=dataset)

    with asection("Estimating region of interest"):
        nb_time_pts = dataset.nb_timepoints(reference_channel)
        slicing = compute_crop_slicing(
            dataset.get_array(reference_channel), [0, nb_time_pts // 2, nb_time_pts - 1], quantile
        )
        aprint("Estimated slicing of", slicing)
        volume_shape = tuple(s.stop - s.start for s in slicing)
        translation = {k: s.start for k, s in zip(("tz", "ty", "tx"), slicing)}
        dest_dataset.append_metadata(translation)

    # Process each channel:
    for channel in dataset._selected_channels(channels):
        with asection(f"Cropping channel {channel}:"):
            array = dataset.get_array(channel)

            dtype = array.dtype
            dest_dataset.add_channel(
                name=channel,
                shape=(len(array),) + volume_shape,
                dtype=dtype,
                chunks=chunks,
                codec=compression,
                clevel=compression_level,
            )

            def process(tp):
                try:
                    aprint(f"Processing time point: {tp} ...")
                    tp_array = array[tp][slicing]
                    dest_dataset.write_stack(channel=channel, time_point=tp, stack_array=tp_array)
                except Exception as error:
                    aprint(error)
                    aprint(f"Error occurred while copying time point {tp} !")
                    import traceback

                    traceback.print_exc()
                    if stop_at_exception:
                        raise error

            if workers == 1:
                for i in range(len(array)):
                    process(i)
            else:
                n_jobs = compute_num_workers(workers, len(array))

                parallel = Parallel(n_jobs=n_jobs)
                parallel(delayed(process)(i) for i in range(len(array)))

    # Dataset info:
    aprint(dest_dataset.info())

    # Check dataset integrity:
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()



# flake8: noqa
from dexp.processing.crop.demo.demo_rep_crop import _demo_representative_crop
from dexp.utils.backends import CupyBackend, NumpyBackend


def test_representative_crop_numpy():
    with NumpyBackend():
        _test_representative_crop()


def test_representative_crop_cupy():
    try:
        with CupyBackend():
            _test_representative_crop()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_representative_crop():
    _demo_representative_crop(display=False, fast_mode=True)
    _demo_representative_crop(display=False, fast_mode=False)



import warnings
from typing import Dict, List, Optional, Tuple

import numpy
import pandas as pd
from arbol import aprint, asection

from dexp.processing.equalise.equalise_intensity import equalise_intensity
from dexp.processing.filters.butterworth_filter import butterworth_filter
from dexp.processing.fusion.dct_fusion import fuse_dct_nd
from dexp.processing.fusion.dft_fusion import fuse_dft_nd
from dexp.processing.fusion.tg_fusion import fuse_tg_nd
from dexp.processing.morphology import area_opening
from dexp.processing.multiview_lightsheet.fusion.basefusion import BaseFusion
from dexp.processing.registration.model import RegistrationModel
from dexp.processing.registration.translation_nd import register_translation_nd
from dexp.processing.registration.translation_nd_proj import (
    register_translation_proj_nd,
)
from dexp.processing.remove_beads import remove_beads_by_threshold
from dexp.processing.restoration.clean_dark_regions import clean_dark_regions
from dexp.processing.restoration.dehazing import dehaze
from dexp.utils import xpArray
from dexp.utils.backends import Backend


class SimViewFusion(BaseFusion):
    def __init__(
        self,
        registration_model: Optional[RegistrationModel],
        equalise: bool,
        equalisation_ratios: List[Optional[float]],
        zero_level: float,
        clip_too_high: int,
        fusion: str,
        fusion_bias_exponent: int,
        fusion_bias_strength_i: float,
        fusion_bias_strength_d: float,
        dehaze_before_fusion: bool,
        dehaze_size: int,
        dehaze_correct_max_level: bool,
        dark_denoise_threshold: int,
        dark_denoise_size: int,
        white_top_hat_size: int,
        white_top_hat_sampling: int,
        remove_beads: bool,
        butterworth_filter_cutoff: float,
        flip_camera1: bool,
        internal_dtype: numpy.dtype = numpy.float16,
        pad: bool = False,
    ):
        super().__init__(
            registration_model,
            equalise,
            equalisation_ratios,
            zero_level,
            clip_too_high,
            fusion,
            dehaze_before_fusion,
            dehaze_size,
            dehaze_correct_max_level,
            dark_denoise_threshold,
            dark_denoise_size,
            butterworth_filter_cutoff,
            internal_dtype,
        )

        self._white_top_hat_size = white_top_hat_size
        self._white_top_hat_sampling = white_top_hat_sampling

        if self._dehaze_size != 0 and self._white_top_hat_size != 0:
            raise ValueError(
                "`dehaze_size` and `white_top_hat_size` cannot be different from"
                f"zero at the same time. Found {self._dehaze_size} and {self._white_top_hat_size}"
            )

        self._fusion_bias_exponent = fusion_bias_exponent
        self._fusion_bias_strength_i = fusion_bias_strength_i
        self._fusion_bias_strength_d = fusion_bias_strength_d
        self._flip_camera1 = flip_camera1
        self._pad = pad
        self._remove_beads = remove_beads

    def _preprocess_single_view(self, view: xpArray, camera: int, lightsheet: int, flip: bool) -> xpArray:
        xp = Backend.get_xp_module()

        with asection(
            f"Moving C{camera}L{lightsheet} and to backend storage and converting to {self._internal_dtype} ..."
        ):
            view = Backend.to_backend(view, dtype=self._internal_dtype, force_copy=False)

        if self._zero_level != 0:
            view = xp.clip(view, self._zero_level, None)
            view -= self._zero_level

        if self._remove_beads:
            with asection(f"Removing beads of C{camera}L{lightsheet}"):
                view = remove_beads_by_threshold(view)

        if self._clip_too_high > 0:
            with asection(f"Clipping intensities above {self._clip_too_high} for C{camera}L{lightsheet}"):
                xp.clip(view, a_min=0, a_max=self._clip_too_high, out=view)

        if self._dehaze_size > 0 and self._dehaze_before_fusion:
            with asection(f"Dehaze C{camera}L{lightsheet} ..."):
                view = dehaze(view, size=self._dehaze_size, minimal_zero_level=0, correct_max_level=True)

        elif self._white_top_hat_size > 0:
            with asection(f"Filtering with White Top Hat transform C{camera}L{lightsheet} ..."):
                view = self._filtered_white_top_hat(view)

        if flip:
            view = xp.flip(view, -1).copy()

        return view

    def _filtered_white_top_hat(self, view: xpArray) -> xpArray:
        # we don't use our area_white_top_hat because we want remove
        # the noise only from the background estimation
        xp = Backend.get_xp_module()
        sp = Backend.get_sp_module()

        # removing salt and pepper noise (morphological opening)
        filtered = sp.ndimage.morphology.grey_opening(view, size=3).astype(numpy.float32)

        # estimating background
        background = area_opening(filtered, self._white_top_hat_size, self._white_top_hat_sampling)

        # white top hat
        view = xp.clip(view - background, a_min=0, a_max=None).astype(view.dtype)

        return view

    def _preprocess_and_fuse_illumination_views(
        self, view0: xpArray, view1: xpArray, flip: bool, camera: int
    ) -> xpArray:

        view0 = self._preprocess_single_view(view0, camera, 0, flip)
        view1 = self._preprocess_single_view(view1, camera, 1, flip)

        if self._equalise:
            with asection(f"Equalise intensity of C{camera}L0 relative to C{camera}L1 ..."):
                view0, view1, ratio = equalise_intensity(
                    view0,
                    view1,
                    zero_level=0,  # already removed when pre-processing single view
                    correction_ratio=self._equalisation_ratios[camera],
                    copy=False,
                )
                aprint(f"Equalisation ratio: {ratio}")
                self._equalisation_ratios[camera] = ratio

        with asection(f"Fuse illumination views C{camera}L0 and C{camera}L1..."):
            fused_view = self._fuse_illumination_views(view0, view1)

        return fused_view

    def preprocess(
        self,
        C0L0: xpArray,
        C0L1: Optional[xpArray] = None,
        C1L0: Optional[xpArray] = None,
        C1L1: Optional[xpArray] = None,
    ) -> Tuple[xpArray, Optional[xpArray]]:

        if C0L1 is not None:
            self._match_input(C0L0, C0L1)
        if C1L0 is not None:
            self._match_input(C0L0, C1L0)
        if C1L1 is not None:
            self._match_input(C0L0, C1L1)

        if C0L1 is None:
            # processing a single light sheet from a single camera
            C0Lx = self._preprocess_single_view(C0L0, camera=0, lightsheet=0, flip=False)
        else:
            C0Lx = self._preprocess_and_fuse_illumination_views(C0L0, C0L1, flip=False, camera=0)

        if C1L0 is None and C1L1 is None:
            # fusing only images from a single camera
            return C0Lx, None

        if C1L1 is None:
            # processing a single light sheet from a single camera
            C1Lx = self._preprocess_single_view(C1L0, camera=1, lightsheet=0, flip=self._flip_camera1)
        else:
            C1Lx = self._preprocess_and_fuse_illumination_views(C1L0, C1L1, flip=self._flip_camera1, camera=1)

        if self._equalise:
            with asection("Equalise intensity of C0Lx relative to C1Lx ..."):
                C0Lx, C1Lx, ratio = equalise_intensity(
                    C0Lx, C1Lx, zero_level=0, correction_ratio=self._equalisation_ratios[2], copy=False
                )
                aprint(f"Equalisation ratio: {ratio}")
                self._equalisation_ratios[2] = ratio

        return C0Lx, C1Lx

    def postprocess(self, CxLx: xpArray) -> xpArray:
        if self._dehaze_size > 0 and not self._dehaze_before_fusion:
            with asection("Dehaze CxLx ..."):
                CxLx = dehaze(
                    CxLx, size=self._dehaze_size, minimal_zero_level=0, correct_max_level=self._dehaze_correct_max_level
                )

        if self._dark_denoise_threshold > 0:
            with asection("Denoise dark regions of CxLx..."):
                CxLx = clean_dark_regions(CxLx, size=self._dark_denoise_size, threshold=self._dark_denoise_threshold)

        if 0 < self._butterworth_filter_cutoff < 1:
            with asection("Filter output using a Butterworth filter"):
                cutoffs = (self._butterworth_filter_cutoff,) * CxLx.ndim
                CxLx = butterworth_filter(CxLx, shape=(31, 31, 31), cutoffs=cutoffs, cutoffs_in_freq_units=False)

        return CxLx

    def fuse(self, C0Lx: xpArray, C1Lx: xpArray, pad: bool) -> xpArray:
        if self._registration_model is None:
            raise RuntimeError("Registration must be computed beforehand.")

        with asection("Register_stacks C0Lx and C1Lx ..."):
            C0Lx, C1Lx = self._registration_model.apply_pair(C0Lx, C1Lx, pad=pad)

        with asection("Fuse detection views C0lx and C1Lx..."):
            CxLx = self._fuse_detection_views(C0Lx, C1Lx)

        return CxLx

    def __call__(
        self,
        C0L0: xpArray,
        C0L1: Optional[xpArray] = None,
        C1L0: Optional[xpArray] = None,
        C1L1: Optional[xpArray] = None,
    ) -> xpArray:

        original_dtype = C0L0.dtype
        xp = Backend.current().get_xp_module()

        C0Lx, C1Lx = self.preprocess(C0L0, C0L1, C1L0, C1L1)
        if C1Lx is None:
            # fusing only two views from two different light sheets
            CxLx = C0Lx
        else:
            CxLx = self.fuse(C0Lx, C1Lx, pad=self._pad)
            CxLx = self.postprocess(CxLx)

        with asection("Converting back to original dtype..."):
            if original_dtype is numpy.uint16:
                CxLx = xp.clip(CxLx, 0, None, out=CxLx)
            CxLx = CxLx.astype(dtype=original_dtype, copy=False)

        return CxLx

    @staticmethod
    def _fuse_views_generic(
        CxL0: xpArray,
        CxL1: xpArray,
        bias_axis: int,
        mode: str,
        smoothing: int,
        bias_exponent: int,
        bias_strength: float,
        downscale: int = 2,
    ) -> xpArray:
        if mode == "tg":
            fused = fuse_tg_nd(
                CxL0,
                CxL1,
                downscale=downscale,
                tenengrad_smoothing=smoothing,
                bias_axis=bias_axis,
                bias_exponent=bias_exponent,
                bias_strength=bias_strength,
            )
        elif mode == "dct":
            fused = fuse_dct_nd(CxL0, CxL1)
        elif mode == "dft":
            fused = fuse_dft_nd(CxL0, CxL1)
        else:
            raise NotImplementedError
        return fused

    def _fuse_illumination_views(self, CxL0: xpArray, CxL1: xpArray, smoothing: int = 12) -> xpArray:
        return self._fuse_views_generic(
            CxL0, CxL1, 2, self._fusion, smoothing, self._fusion_bias_exponent, self._fusion_bias_strength_i
        )

    def _fuse_detection_views(self, C0Lx: xpArray, C1Lx: xpArray, smoothing: int = 12) -> xpArray:
        return self._fuse_views_generic(
            C0Lx, C1Lx, 0, self._fusion, smoothing, self._fusion_bias_exponent, self._fusion_bias_strength_d
        )

    def compute_registration(
        self, C0Lx: xpArray, C1Lx: xpArray, mode: str, edge_filter: bool, crop_factor_along_z: float
    ) -> None:
        C0Lx = Backend.to_backend(C0Lx)
        C1Lx = Backend.to_backend(C1Lx)

        depth = C0Lx.shape[0]
        crop = int(depth * crop_factor_along_z)
        C0Lx_c = C0Lx[crop:-crop]
        C1Lx_c = C1Lx[crop:-crop]

        if mode == "projection":
            self._registration_model = register_translation_proj_nd(C0Lx_c, C1Lx_c, edge_filter=edge_filter)
        elif mode == "full":
            self._registration_model = register_translation_nd(C0Lx_c, C1Lx_c, edge_filter=edge_filter)
        else:
            raise NotImplementedError

        aprint(
            f"Applying registration model: {self._registration_model},"
            f"overall confidence: {self._registration_model.overall_confidence()}"
        )

    @staticmethod
    def validate_views(views: Dict[str, xpArray]) -> Dict[str, xpArray]:
        if not isinstance(views, dict):
            raise ValueError("Views must be of type Dict[str, xpArray]")

        if len(views) == 4:
            return views
        assert len(views) == 2, f"Found only {views.keys()}, expected 2 or 4 views."
        warnings.warn(f"Only two views found. Fusing according to channels' suffixes {views.keys()}")

        if all("C1" in k for k in views):
            warnings.warn("Swapping names from C1 to C0")
            return {k.replace("C1", "C0"): v for k, v in views.items()}

        if all("L1" in k for k in views):
            warnings.warn("Swapping names from L1 to L0")
            return {k.replace("L1", "L0"): v for k, v in views.items()}

        return views


def summary_from_simview_models(models: List[SimViewFusion]) -> pd.DataFrame:
    df = []
    for m in models:
        current = {}
        for i, eq_ratio in enumerate(m._equalisation_ratios):
            if eq_ratio is not None:
                current[f"eq_ratio_{i}"] = eq_ratio

        if m.registration_model is not None:
            for i, v in enumerate(m.registration_model.shift_vector):
                current[f"shift_{i}"] = v
            current["confidence"] = m.registration_model.overall_confidence()

    return pd.DataFrame(df)



from dexp.processing.denoising.demo.demo_2D_gaussian import _demo_gaussian
from dexp.utils.testing import execute_both_backends


@execute_both_backends
def test_gaussian():
    assert _demo_gaussian(display=False) >= 0.600 - 0.02



import click
from arbol.arbol import aprint

from dexp.cli.defaults import DEFAULT_WORKERS_BACKEND
from dexp.video.blend import blend_color_image_sequences


@click.command()
@click.argument("input_paths", nargs=-1)
@click.option("--output_path", "-o", type=str, default=None, help="Output folder for blended frames.")
@click.option(
    "--modes",
    "-m",
    type=str,
    default="max",
    help="Either one for all images, or one per image in the form of a sequence. Blending modes are: mean, add, satadd, max, alpha. In the add mode the values are clipped. For alpha-blending, we assume that RGB values are not ‘alpha-premultiplied’.",
    show_default=True,
)
@click.option(
    "--alphas", "-a", type=str, default="1.0", help="List of alpha values for each input image", show_default=True
)
@click.option(
    "--scales",
    "-s",
    type=str,
    default="1.0",
    help="List of scales ‘s’ for each input image, starting with the second one (first image has always a scale of 1). Scaled images must be of same dimension or smaller than the first images. ",
    show_default=True,
)
@click.option(
    "--translations",
    "-t",
    type=str,
    default="0,0",
    help="List of translations ‘x0,y0|x1_y1|...’ for each input image, starting with the second one (first image is always at 0,0). If set to 0,0 no translation is performed for any image. ",
    show_default=True,
)
@click.option("--borderwidth", "-bw", type=int, default=0, help="Border width added to insets. ", show_default=True)
@click.option(
    "--bordercolor",
    "-bc",
    type=str,
    default="1,1,1,1",
    help="Border color in RGBA format. For example: 1,1,1,1 is white.",
    show_default=True,
)
@click.option(
    "--borderover",
    "-bo",
    is_flag=True,
    help="Border is overlayed on top of inset images, without increasing their size. ",
    show_default=True,
)
@click.option("--overwrite", "-w", is_flag=True, help="Force overwrite of output images.", show_default=True)
@click.option(
    "--workers",
    "-k",
    type=int,
    default=-1,
    help="Number of worker threads to spawn, set to -1 for maximum number of workers",
    show_default=True,
)  #
@click.option(
    "--workersbackend",
    "-wkb",
    type=str,
    default=DEFAULT_WORKERS_BACKEND,
    help="What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ",
    show_default=True,
)  #
@click.option("--device", "-d", type=int, default=0, help="Sets the CUDA devices id, e.g. 0,1,2", show_default=True)  #
def blend(
    input_paths,
    output_path,
    modes,
    alphas,
    scales,
    translations,
    borderwidth,
    bordercolor,
    borderover,
    overwrite,
    workers,
    workersbackend,
    device,
):
    """Blends videos or still images together, using different alphas, blend modes, and possibly with rescaling, translation, and added borders."""

    number_of_inputs = len(input_paths)

    if output_path is None:
        output_path = input_paths[0] + "_blend"
    elif output_path.startswith("_"):
        output_path = input_paths[0] + output_path

    if "," in modes:
        modes = tuple(mode.strip() for mode in modes.split(","))
    else:
        modes = (modes,) * number_of_inputs

    if "," in alphas:
        alphas = tuple(float(alpha) for alpha in alphas.split(","))
    else:
        alphas = (float(alphas),) * number_of_inputs

    if "," in scales:
        scales = tuple(float(scale) for scale in scales.split(","))
    else:
        scales = (float(scales),) * number_of_inputs

    if "|" in translations:
        translations = tuple(tuple(float(v) for v in xy.split(",")) for xy in translations.split("|"))
    elif "," in translations:
        translations = (tuple(float(v) for v in translations.split(",")),) * number_of_inputs
    else:
        translations = ((0, 0),) * number_of_inputs

    bordercolor = tuple(float(v) for v in bordercolor.split(","))

    if len(input_paths) <= 1:
        aprint("Blending requires at least two paths!")
        return

    blend_color_image_sequences(
        input_paths=input_paths,
        output_path=output_path,
        modes=modes,
        alphas=alphas,
        scales=scales,
        translations=translations,
        border_width=borderwidth,
        border_color=bordercolor,
        border_over_image=borderover,
        overwrite=overwrite,
        workers=workers,
        workersbackend=workersbackend,
        device=device,
    )



from typing import Tuple


def get_3d_image_graph(shape: Tuple[int, ...]):
    assert len(shape) == 3
    import higra as hg

    neighbors = []
    for i in range(3):
        for j in (1, -1):
            shift = [0] * 3
            shift[i] = j
            neighbors.append(shift)

    return hg.get_nd_regular_graph(shape, neighbors)



from contextlib import contextmanager
from time import time

from arbol import aprint


@contextmanager
def timeit(description: str = "") -> None:
    start = time()
    yield
    elapsed_time = time() - start

    aprint(f"Elapsed time for '{description}': {elapsed_time} seconds")
    return elapsed_time



import json
from typing import Sequence, Union

from arbol import asection

from dexp.processing.registration.model.pairwise_registration_model import (
    PairwiseRegistrationModel,
)
from dexp.processing.registration.model.sequence_registration_model import (
    SequenceRegistrationModel,
)
from dexp.processing.registration.model.translation_registration_model import (
    TranslationRegistrationModel,
)
from dexp.processing.registration.model.warp_registration_model import (
    WarpRegistrationModel,
)
from dexp.utils.backends import Backend


def from_json(json_str: str) -> Union[PairwiseRegistrationModel, SequenceRegistrationModel]:
    xp = Backend.get_xp_module()
    parsed_model = json.loads(json_str)
    if parsed_model["type"] == "translation":
        model = TranslationRegistrationModel(
            shift_vector=xp.asarray(parsed_model["translation"]), confidence=xp.asarray(parsed_model["confidence"])
        )
    elif parsed_model["type"] == "warp":
        model = WarpRegistrationModel(
            vector_field=xp.asarray(parsed_model["vector_field"]), confidence=xp.asarray(parsed_model["confidence"])
        )

    elif parsed_model["type"] == "translation_sequence":
        model_list_json = parsed_model["model_list"]
        model_list = list(from_json(model_json) for model_json in model_list_json)
        model = SequenceRegistrationModel(model_list=model_list)

    else:
        raise NotImplementedError

    return model


def model_list_from_file(file_path: str) -> Sequence[Union[PairwiseRegistrationModel, SequenceRegistrationModel]]:
    with open(file_path) as models_file:
        lines = models_file.readlines()

        model_list = []
        with asection(f"Loading {len(lines)} models from file: {file_path}"):
            for line in lines:
                line = line.strip()
                model = None if line == "None" else from_json(line)
                model_list.append(model)

        return model_list


def model_list_to_file(file_path: str, model_list: Sequence[PairwiseRegistrationModel]):
    with open(file_path, "w") as models_file:
        lines = []

        with asection(f"Writing {len(model_list)} models to file: {file_path}"):
            for model in model_list:
                line = "None" if model is None else model.to_json()
                lines.append(line + "\n")

            models_file.writelines(lines)



import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import DEFAULT_WORKERS_BACKEND
from dexp.cli.parsing import _get_output_path, _parse_channels, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.tiff import dataset_tiff


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option("--channels", "-c", default=None, help="selected channels.")  #
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)  #
@click.option(
    "--overwrite", "-w", is_flag=True, help="to force overwrite of target", show_default=True
)  # , help='dataset slice'
@click.option(
    "--project", "-p", type=int, default=None, help="max projection over given axis (0->T, 1->Z, 2->Y, 3->X)"
)  # , help='dataset slice'
@click.option(
    "--split",
    is_flag=True,
    help="Splits dataset along first dimension, be carefull, if you slice to a single time point this will split along z!",
)  # , help='dataset slice'
@click.option(
    "--clevel", "-l", type=int, default=0, help="Compression level, 0 means no compression, max is 9", show_default=True
)  # , help='dataset slice'
@click.option(
    "--workers",
    "-k",
    default=-4,
    help="Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n|",
    show_default=True,
)  #
@click.option(
    "--workersbackend",
    "-wkb",
    type=str,
    default=DEFAULT_WORKERS_BACKEND,
    help="What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ",
    show_default=True,
)
def tiff(input_paths, output_path, channels, slicing, overwrite, project, split, clevel, workers, workersbackend):
    """Exports dataset as TIFF file(s)."""

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path)
    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)

    with asection(
        f"Exporting to TIFF datset: {input_paths}, channels: {channels}, slice: {slicing}, project:{project}, split:{split}"
    ):
        dataset_tiff(
            input_dataset,
            output_path,
            channels=channels,
            slicing=slicing,
            overwrite=overwrite,
            project=project,
            one_file_per_first_dim=split,
            clevel=clevel,
            workers=workers,
            workersbackend=workersbackend,
        )

        input_dataset.close()
        aprint("Done!")
